/*
 * The steady-state of a 2LS subject to a strong driving field: Find the steady
 * state by finding the eigenstates of the propagator.
 *
 * Robert Johansson <robert@riken.jp>
 *
 * example run:
 *
 * $ ./run_2ls_driven_pss 0.0 0.0 0.1 5.0
 *  
 */
#include "steadystate.h"
#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "hamiltonian.h"
#include "master_equation.h"
#include "gsl_ext.h"
#include "param.h"
#include "misc.h"

#include <math.h>
#include <stdio.h> 
#include <strings.h> 
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>

//#define DATADIR_TEMPLATE "/home/rob/qdpack-data/%s"
#define DATADIR_TEMPLATE "data/%s"

static char DATADIR[256]; 
static double pe_cb_sum = 0.0, pe_eb_sum = 0.0;
static int    pe_no  = 0;
static gsl_matrix_complex *rho_eb_t, *rho_tmp_t;

#define USE_EIGENBASIS 1

/**
 * Strong driving hamiltonian
 *
 */
int
hamiltonian_sd_t(gsl_matrix_complex *H_t, gsl_matrix_complex *H0, gsl_matrix_complex *H1, double t, solver_param *p)
{
    if (p->f_A == 0.0)
    {
        gsl_matrix_complex_memcpy(H_t, H0);
    }
    else
    {
        gsl_complex z;

		// H_t = H0 + f(t) H1
        // f(t) = A * sin(w * t)
        gsl_matrix_complex_memcpy(H_t, H1);
        GSL_SET_COMPLEX(&z, p->f_A * sin(p->f_w * t) / 2, 0.0);
        gsl_matrix_complex_scale(H_t, z);
        gsl_matrix_complex_add(H_t, H0);
    }

    return 0;
}


/* ---
 * Process the density matrix for time t=T_final (calc. expectations values, 
 * store to file, etc.) 
 */
int
rho_store_final_cb(gsl_matrix_complex *rho_cb_t, double t, quantum_system *qs, solver_param *sp)
{
	char filename[128], row[1024];
	gsl_matrix_complex *dm_part;
	
	sp->Ht_func(sp->H_t, sp->H0, sp->H1, sp->Tf, sp);

	if (USE_EIGENBASIS == 1)
	{
		gsl_eigen_hermv(sp->H_t, sp->eval, sp->evec, sp->w);
		gsl_eigen_hermv_sort(sp->eval, sp->evec, GSL_EIGEN_SORT_VAL_ASC); 

		// rho_eb[i] = S^{-1} rho_cb S
		gsl_blas_zgemm(CblasConjTrans, CblasNoTrans, GSL_COMPLEX_ONE, sp->evec,  rho_cb_t, GSL_COMPLEX_ZERO, rho_tmp_t);
		gsl_blas_zgemm(CblasNoTrans,   CblasNoTrans, GSL_COMPLEX_ONE, rho_tmp_t, sp->evec, GSL_COMPLEX_ZERO, rho_eb_t);
	}

    //
    // store the qubit excitation probability in both charge basis (rho_cb_t) 
    // and eigenbasis (rho_eb_t)
    //
	snprintf(filename, sizeof(filename), "%s/p_e%s.dat", DATADIR, sp->simsig);
	snprintf(row, sizeof(row), "%f\t%f\t%f\t%f",
		     sp->x, sp->y, 
		     GSL_REAL(gsl_matrix_complex_get(rho_cb_t, 1, 1)),
		     GSL_REAL(gsl_matrix_complex_get(rho_eb_t, 1, 1)));

//	snprintf(row, sizeof(row), "%f\t%f\t%f\t%f\t%f\t%f\t%d",
//		     sp->x, sp->y, 
//		     GSL_REAL(gsl_matrix_complex_get(rho_cb_t, 1, 1)),
//		     GSL_REAL(gsl_matrix_complex_get(rho_eb_t, 1, 1)),
// 		     pe_eb_sum / pe_no, pe_cb_sum / pe_no, pe_no);

	misc_file_row_add(filename, row);		

	return 0;
}
	
/* ---
 * Process the density matrix for time t (calc. expectations values, 
 * store to file, etc.) 
 */
int
rho_store_cb(gsl_matrix_complex *rho_t, double t, quantum_system *qs, solver_param *sp)
{
	if (t == sp->Tf)
	{
	 	gsl_matrix_complex_memcpy(sp->rho_f, rho_t);
	}
	
	return 0;
}

/* ---
 * Process the density matrix for time t (average excitation probability)
 */
int
rho_average_cb(gsl_matrix_complex *rho_cb_t, double t, quantum_system *qs, solver_param *sp)
{
	if (USE_EIGENBASIS == 1)
	{
        // rho_eb[i] = S^{-1} rho_cb S
		gsl_eigen_hermv(sp->H_t, sp->eval, sp->evec, sp->w);
		gsl_eigen_hermv_sort(sp->eval, sp->evec, GSL_EIGEN_SORT_VAL_ASC); 

		gsl_blas_zgemm(CblasConjTrans, CblasNoTrans, GSL_COMPLEX_ONE, sp->evec,  rho_cb_t, GSL_COMPLEX_ZERO, rho_tmp_t);
		gsl_blas_zgemm(CblasNoTrans,   CblasNoTrans, GSL_COMPLEX_ONE, rho_tmp_t, sp->evec, GSL_COMPLEX_ZERO, rho_eb_t);
	}

	if (1)
	{
        pe_cb_sum += GSL_REAL(gsl_matrix_complex_get(rho_cb_t, 1, 1));
		pe_eb_sum += GSL_REAL(gsl_matrix_complex_get(rho_eb_t, 1, 1));
		pe_no++;
	}
	
	
	return 0;
}

/* ---
 * Simulation parameters.
 * 
 */
double epsilon[] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double delta[]   = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double ho_w[]    = { 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double *lambda[10];


/* ---
 * Program starts here.
 * 
 */
int
main(int argc, char **argv)
{
	quantum_system *qs;
	gsl_matrix_complex *rho_q, *rho_c, *rho_ss;
	gsl_matrix_complex *rho0, *rho_t, *U;
	density_matrix_list_t dm_list;
	int qsn, i, j, noff, nsize, npeak;
	double Navg, fw, fA, phi, Gc, Gq, gqc, x, y, pDelta, pEpsilon, G1, G2, T1, T2;
	double yi, yf, yd;

	solver_param param;
	bzero((void *)&param, sizeof(solver_param));

	if (argc != 5)
	{
		printf("usage: %s x y-init y-delta y-final\n", argv[0]);
		exit(0);
	}

	snprintf(DATADIR, sizeof(DATADIR), DATADIR_TEMPLATE, argv[0]);

    // ---- PARAMETERS START

    //
    // frequency in units of GHz and time in units of ns
    //

    //
    // x = coefficient of sigma_z in qubit hamiltonian, in units of driving frequency
    // x = epsilon / (fw * 2 * M_PI)
    //
    // y = the driving amplitude in units of driving frequency
    // y = f_A / (fw * 2 * M_PI)
    //

    // get values from command line parameters
	x      = strtod(argv[1], NULL);
	yi     = strtod(argv[2], NULL);
	yd     = strtod(argv[3], NULL);
	yf     = strtod(argv[4], NULL);
	
    fw     = 1.0;               // driving frequency
    pDelta  = fw/300.0;         // coefficient of sigma_x in qubit hamiltonain
    T1     = 100000000.0 / fw;       // relaxation time
    T2     =      500.0 / fw;       // dephasing  time

    // ---- PARAMETERS END
   
	param.epsilon = epsilon;
	param.delta   = delta;
	param.ho_w    = ho_w;
	param.x = x;
	for (i = 0; i < 10; i++)
	{
		param.epsilon[i] *= 2*M_PI;
		param.delta[i]   *= 2*M_PI;
		param.ho_w[i]    *= 2*M_PI;
		lambda[i]         = (double *)calloc(10, sizeof(double));
	}
	param.lambda = lambda;
    G1 = 1.0/T1;
    G2 = 1.0/T2;
	param.cont_basis_transform = 0;

	param.f_w = fw * 2 * M_PI;

	param.H0_func = (hamiltonian_func_t)hamiltonian_nTLS;
	param.H1_func = (hamiltonian_func_t)hamiltonian_z_drive;
	param.Ht_func = (hamiltonian_td_func_t)hamiltonian_sd_t;

	/* 
	 * create a new quantum system object
	 */
	qs = quantum_system_new();
	quantum_system_add(qs, 2);        // a 2-level system
	qsn = quantum_system_nstates(qs);
	param.qs = qs;


	/*
	 * Set up dissipation
	 *
	 */
	//param.wT = 0.0 * 2 * M_PI;
	param.do_n = 0;
	Navg = 0;                       // temperature zero kelvin

	// relaxation
	param.do_a[param.do_n]  = gsl_matrix_complex_alloc(qsn, qsn);
	operator_ho_lowering(param.do_a[param.do_n], qs, 0, 0); 
	param.do_ad[param.do_n]	= gsl_matrix_complex_alloc(qsn, qsn); 
	operator_ho_raising(param.do_ad[param.do_n], qs, 0, 0);
	param.do_g1[param.do_n] = G1 * (1 + Navg); // relaxation
	param.do_g2[param.do_n] = G1 * Navg;       // excitation
	param.do_n++;

	// dephasing
	param.do_a[param.do_n]  = gsl_matrix_complex_alloc(qsn, qsn);
	operator_sigma_z(param.do_a[param.do_n], qs, 0);
    param.do_ad[param.do_n] = gsl_matrix_complex_alloc(qsn, qsn);
	operator_sigma_z(param.do_ad[param.do_n], qs, 0); 
    param.do_g1[param.do_n] = G2 / 2; 
    param.do_g2[param.do_n] = G2 / 2; 
    param.do_n++;

    // pre-allocate memory for diagonalizations
	param.eval = gsl_vector_alloc(qsn);
    param.evec = gsl_matrix_complex_alloc(qsn, qsn);
	param.w    = gsl_eigen_hermv_alloc(qsn);

    // output data file name signature
	snprintf(param.simsig, sizeof(param.simsig), "_dDelta_%.3f_G1_%.6f_G2_%.4f_fw_%.3f", pDelta, G1, G2, fw);

	/* storage space for final rho */
	param.rho_f = gsl_matrix_complex_alloc(qsn, qsn);
	rho_eb_t    = gsl_matrix_complex_alloc(qsn, qsn);
    gsl_matrix_complex_set_zero(rho_eb_t);
	rho_tmp_t  = gsl_matrix_complex_alloc(qsn, qsn);
	rho_ss     = gsl_matrix_complex_alloc(qsn, qsn);

	/*
	 *  allocate matrices for hamiltonians and the propagator
	 */
	U = gsl_matrix_complex_alloc(qsn*qsn, qsn*qsn);
    param.H_t = gsl_matrix_complex_alloc(qsn, qsn);
    param.H0 = gsl_matrix_complex_alloc(qsn, qsn);
    param.H1 = gsl_matrix_complex_alloc(qsn, qsn);

	/*
	 * Setup initial state
	 */
	rho_q = density_matrix_pure_TLS(0.0); // start in qubit ground state
    density_matrix_list_init(&dm_list);
    density_matrix_list_append(&dm_list, rho_q);
	rho0 = gsl_matrix_complex_alloc(qsn, qsn);
	density_matrix_combine(rho0, qs, &dm_list);


    //
    // start sweep of parameter range
    //
	delta[0]   = pDelta * (2*M_PI);     // qubit parameter: sigma_x coefficient
	epsilon[0] = x * fw * (2*M_PI);     // qubit parameter: sigma_y coefficient

	for (y = yi; y <= yf; y += yd)
	{	
  		pe_cb_sum = 0.0;
        pe_eb_sum = 0.0;
		pe_no  = 0;

        // sweep driving amplitude
		param.y = y;	
	    param.f_A  = y * fw * 2 * M_PI;

    	printf("ITERATION: %s: x = %f, y = %f\n", param.simsig, param.x, param.y);


		/*
		 * calculate propagator:
		 *
		 */
		param.Ti = 0.0;
		param.Tf = 1/fw;
		param.dT = 1/fw;
		if (dm_calculate_propagator(qs, &param, rho_store_cb, U) == -1)
		{
			fprintf(stderr, "Evaluation of the propagator failed.\n");
			return -1;
		}

		/*
		 * Find the steady state from the evolution operator
		 *
		 */
		if (dm_steady_state_from_propagator(qs, &param, rho_ss, U) == -1)
		{
			fprintf(stderr, "Evaluation of the propagator failed.\n");
			return -1;
		}

        //
        // optionally evolve the density matrix for one driving period
        // and average
        // param.dT = 0.01/fw; // one period
        // if (dm_evolve_lme_ib_t(qs, rho_ss, &param, rho_average_cb) != 0)
        // {
        //	  fprintf(stderr, "Evolution of the quantum system failed.\n");
        //    return -1;
        // }
	
		rho_store_final_cb(rho_ss, param.Tf, qs, &param);
	}

	return 0;

}



