/*
 * Functions for managing basis transformations.
 * 
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: basis_transform.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */

#include <stdio.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

#include <gsl/gsl_eigen.h>
#include <gsl/gsl_sf_legendre.h>

#include "quantum_system.h"
#include "density_matrix.h"
#include "basis_transform.h"
#include "operators.h"
#include "misc.h"
#include "gsl_ext.h"
#include "mlib.h"
#include "param.h"


/*
 * Generate basis transformation matrix for a harmonic oscillator with
 * frequency switch: S
 *
 * rho_t = S * rho_t0 * S'
 *
 */
gsl_matrix_complex *
basis_transform(quantum_system *qs, solver_param *param, int n, ho_w_func_t ho_w_cb, double t)
{
	gsl_complex z;
	gsl_matrix_complex *S;
	int i, j, ns, mm, ll;
	double r, w0, wt, sij;
	density_matrix_list_t op_list;

	ns = qs->nstates[n];

	if ((S = gsl_matrix_complex_alloc(ns,ns)) == NULL)
	{
		fprintf(stderr, "basis_transform_ho: Failed to allocate basis tranformation matrix (Out of memory?).\n");
	        return NULL;
	}
	gsl_matrix_complex_set_zero(S);
	
	w0 = ho_w_cb(0.0, param);
	wt = ho_w_cb(t, param);	

	r = 2 * sqrt(w0 * wt) / (wt + w0);
	if (r > 1.0)
		r = 1.0;
	else if (r < -1.0)
		r = -1.0;
	
	//printf("DEBUG: basis_transform: r = %f (%f, %f): %f\n", r, w0, wt, t);
	
	for (j = 0; j < ns; j++)
	{
		for (i = j; i < ns; i++)
		{
			if ((i == j) || ((i+j)%2 == 0))
			{
				ll = (i+j)/2;
				mm = abs(i-j)/2;
				sij = fact_sqrt(MIN(i,j)) / fact_sqrt(MAX(i,j)) * sqrt(r) * gsl_sf_legendre_Plm(ll, mm, r);
				//if (wt > w0)
				GSL_SET_COMPLEX(&z, sij, 0.0);
				//else	
				//	GSL_SET_COMPLEX(&z, -sij, 0.0);
				gsl_matrix_complex_set(S, i, j, z);
				
				if (i != j)
				{
					sij = pow(-1.0, (j-i)/2.0) * sij;
					GSL_SET_COMPLEX(&z, sij, 0.0);
					gsl_matrix_complex_set(S, j, i, z);
				}
			}
		}
	}

	if (qs->nsubsys > 1)
	{
        	density_matrix_list_init(&op_list);

        	for (i = 0; i < qs->nsubsys; i++)
        	{
        	        gsl_matrix_complex *m = gsl_matrix_complex_alloc(qs->nstates[i],qs->nstates[i]);

        	        if (i != n)
               		{
                        	operator_unit(m, qs, i);
                	}
                	else
                	{
				m = S;
                	}

                	density_matrix_list_append(&op_list, m);
        	}
        	density_matrix_combine(S, qs, &op_list);
	        density_matrix_list_free_entries(&op_list);
	}
	
	return S;
}

/*
 * Calculate dS/dt at time t. 
 * 
 * Method: dS/dt = (S(t+h)-S(t-h))/(2h)
 */
gsl_matrix_complex *
basis_transform_derivative(quantum_system *qs, solver_param *param, int n, ho_w_func_t ho_w_cb, double t)
{	
	double h;
	gsl_complex z;
	gsl_matrix_complex *S1, *S2;

	h = 1e-6;
	S1 = basis_transform(qs, param, n, ho_w_cb, t-h);
	S2 = basis_transform(qs, param, n, ho_w_cb, t+h);

	gsl_matrix_complex_sub(S2, S1);
	
	if (ho_w_cb(t, param) < ho_w_cb(0.0, param))
		GSL_SET_COMPLEX(&z, -1.0 / (2.0 * h), 0.0);
	else
		GSL_SET_COMPLEX(&z, 1.0 / (2.0 * h), 0.0);
		
	gsl_matrix_complex_scale(S2, z);

	gsl_matrix_complex_free(S1);

	return S2;
}	



