/*
 * Functions for managing density matrices.
 * 
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: density_matrix.c,v 1.4 2008-09-24 10:14:52 rob Exp $
 */

#include <stdio.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

#include <gsl/gsl_eigen.h>


#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "misc.h"
#include "gsl_ext.h"
#include "mlib.h"

/**
 * \breif Tensor productor of density matrices
 *
 * Combine a list of density matrices for the sub-systems into
 * a density matrix for the full system (product state), by taking
 * the tensor product.
 *
 * @param	qs		Data structure that defines the quantum system
 * @param	dm_list		A list of density matrices that are to be tensor multiplied
 *
 */
int 
density_matrix_combine(gsl_matrix_complex *dm_full, quantum_system *qs, density_matrix_list_t *dm_list)
{
	int ns, i, j, k;
//	gsl_matrix_complex *dm_full;
	gsl_complex dm_ij;
	quantum_system_state_vector_t qsv1, qsv2;

	ns = quantum_system_nstates(qs);

	if (qs == NULL || dm_list == NULL || qs->nsubsys != dm_list->len)
	{
		fprintf(stderr, "density_matrix_combine: NULL arguments or system-size/dm_list mismatch.\n");
		return -1;
	}
	
//	if ((dm_full = gsl_matrix_complex_alloc(ns,ns)) == NULL)
//	{
//		fprintf(stderr, "density_matrix_combine: Failed to allocate dm_full (Out of memory?).\n");
//		return NULL;
//	}

	for (i = 0; i < ns; i++)
	{
		quantum_system_state_vector(qs, i, qsv1);
		
		for (j = 0; j < ns; j++)
		{
			quantum_system_state_vector(qs, j, qsv2);
			
			GSL_SET_COMPLEX(&dm_ij, 1.0, 0.0);
			
			for (k = 0; k < qs->nsubsys; k++)
			{
				dm_ij = gsl_complex_mul(dm_ij, gsl_matrix_complex_get(dm_list->dm[k],qsv1[k],qsv2[k]));
			}
			gsl_matrix_complex_set(dm_full, i, j, dm_ij);
		}
	}
	
	return 0;
}

/**
 * \breif	Trace out all sussystem except the nth system
 *
 * Trace out all sub-system except sub-system n from the full-system
 * density matrix dm.
 *
 * @param	qs              Data structure that defines the quantum system
 * @param	dm		Full system density matrix
 * @param	n		The subsystem that is to be kept (all other subsystem traced out)
 *
 */
gsl_matrix_complex *
density_matrix_traceout(quantum_system *qs, gsl_matrix_complex *dm, int n)
{
	int i, j, ii, jj, k, ns_part, ns_sub;
	gsl_matrix_complex *dm_part;
	gsl_complex dm_ij;
	quantum_system_state_vector_t qsv1, qsv2;
	quantum_system *qs_sub;	

	ns_part = qs->nstates[n];
	if ((dm_part = gsl_matrix_complex_alloc(ns_part,ns_part)) == NULL)
	{
		fprintf(stderr, "density_matrix_traceout: Failed to allocate dm_part (Out of memory?).\n");
		return NULL;
	}

	qs_sub = quantum_system_copy(qs);
	qs_sub->nstates[n] = 1;
	ns_sub = quantum_system_nstates(qs_sub);
	
	//printf("DEBUG: ns = %d, ns_part = %d, ns_sub = %d\n", quantum_system_nstates(qs), ns_part, ns_sub);
	for (i = 0; i < ns_part; i++)
	{
		for (j = 0; j < ns_part; j++)
		{
			GSL_SET_COMPLEX(&dm_ij, 0.0, 0.0);	
			for (k = 0; k < ns_sub; k++)
			{
				quantum_system_state_vector(qs_sub, k, qsv2);
				quantum_system_state_vector(qs_sub, k, qsv1);
			
				qsv1[n] = i;
				qsv2[n] = j;

				ii = quantum_system_state_number(qs, qsv1);
				jj = quantum_system_state_number(qs, qsv2);
			
				//printf("DEBUG: k = %d, ii = %d, jj = %d [%1.2f]\n", k, ii, jj, GSL_REAL(gsl_matrix_complex_get(dm, ii, jj)));
				//printf("DEBUG: %d\t", ii); quantum_system_state_vector_print(qs, qsv1); printf("\n");
				//printf("DEBUG: %d\t", jj); quantum_system_state_vector_print(qs, qsv2); printf("\n");

				dm_ij = gsl_complex_add(dm_ij, gsl_matrix_complex_get(dm, ii, jj));
			}

			gsl_matrix_complex_set(dm_part, i, j, dm_ij);
		}
	}
		
	return dm_part;
}

/**
 * \breif	Trace out a few of the subsystems from the full density matrix.
 *
 * Trace out subsystems marked in mask variable from the full-system density
 * matrix dm. 
 *
 * mask[n] = 1: -> trace out subsystem n
 * mask[n] = 0: -> keep subsystem n
 *
 * qs_sub:   part of system to be kept
 * qs_trace: part of system to be traced
 *
 * @param       qs        Data structure that defines the quantum system
 * @param       dm        Full system density matrix
 * @param      	mask	  Mask of subsystems to keep/trace out, see above.
 *
 */
gsl_matrix_complex *
density_matrix_traceout_multi(quantum_system *qs, gsl_matrix_complex *dm, quantum_system *mask)
{
	int i, j, ii, jj, k, ns_trace, ns_sub, n, m;
	gsl_matrix_complex *dm_sub;
	gsl_complex dm_ij;
	quantum_system_state_vector_t qsv_trace_i, qsv_trace_j, qsv_sub_i, qsv_sub_j;
	quantum_system *qs_sub, *qs_trace;	

	qs_trace = quantum_system_copy(qs);
	qs_sub   = quantum_system_new();
	for (n = 0; n < mask->nsubsys; n++)
	{
		if (mask->nstates[n] == 0)
		{	
			// If subsystem is masked "0", add it to qs_sub and remove from qs_trace
			quantum_system_add(qs_sub, qs->nstates[n]);
			qs_trace->nstates[n] = 1;
		}
	}
	ns_sub   = quantum_system_nstates(qs_sub);
	ns_trace = quantum_system_nstates(qs_trace);

	if ((dm_sub = gsl_matrix_complex_alloc(ns_sub,ns_sub)) == NULL)
	{
		fprintf(stderr, "density_matrix_traceout: Failed to allocate dm_part (Out of memory?).\n");
		return NULL;
	}
	
	for (i = 0; i < ns_sub; i++)
	{
		quantum_system_state_vector(qs_sub, i, qsv_sub_i);
		for (j = 0; j < ns_sub; j++)
		{
			quantum_system_state_vector(qs_sub, j, qsv_sub_j);
			
			GSL_SET_COMPLEX(&dm_ij, 0.0, 0.0);	
			for (k = 0; k < ns_trace; k++)
			{
				quantum_system_state_vector(qs_trace, k, qsv_trace_i);
				quantum_system_state_vector(qs_trace, k, qsv_trace_j);
			
				m = 0;
				for (n = 0; n < mask->nsubsys; n++)
				{
					if (mask->nstates[n] == 0)
					{
						qsv_trace_i[n] = qsv_sub_i[m];
						qsv_trace_j[n] = qsv_sub_j[m];
						m++;
					}				
				}
				
				ii = quantum_system_state_number(qs, qsv_trace_i);
				jj = quantum_system_state_number(qs, qsv_trace_j);
			
				dm_ij = gsl_complex_add(dm_ij, gsl_matrix_complex_get(dm, ii, jj));
			}

			gsl_matrix_complex_set(dm_sub, i, j, dm_ij);
		}
	}
		
	return dm_sub;
}


/**
 * Initialize the data structure that describe a list of 
 * density matrices.
 *
 * @param	dm_list		List of density matrices to initiate.
 */
void
density_matrix_list_init(density_matrix_list_t *dm_list)
{
	int i;
	
	if (dm_list != NULL)
	{
		for (i = 0; i < NQS_MAX; i++)
			dm_list->dm[i] = NULL;
		dm_list->len = 0;
	}
}

/**
 * Append a density matrix to list of density matrices that describe 
 * sub-systems of the full quantum system.
 *
 * @param	dm_list		List of density matrices.
 * @param	dm		Density matrix of subsystem to add to the list.
 */
void
density_matrix_list_append(density_matrix_list_t *dm_list, gsl_matrix_complex *dm)
{
	if (dm_list->len < NQS_MAX)
		dm_list->dm[dm_list->len++] = dm;
	else
		fprintf(stderr, "density_matrix_list_append: Reached max size.\n");
}

/**
 * Free density matrices in the list. 
 * 
 * @param	dm_list		List of density matrices.
 */
void
density_matrix_list_free_entries(density_matrix_list_t *dm_list)
{
	int i;

	for (i = 0; i < dm_list->len; i++)
		gsl_matrix_complex_free(dm_list->dm[i]);
		
//		free(dm_list->dm[i]);
}


/**
 * \breif	Initial density matrix for a two-level systems.
 *
 * Generate a pure-state density matrix for a TLS. (Used for constructing
 * initial states).
 *
 * p_ex = probability of being in the excited state
 *
 * @param	p_ex	Occupation probability of the two-level system excited state.
 */
gsl_matrix_complex *
density_matrix_pure_TLS(double p_ex)
{
	double aa, bb;
	gsl_matrix_complex *dm;
	gsl_complex z;
	
        if ((dm = gsl_matrix_complex_alloc(2,2)) == NULL)
        {
                fprintf(stderr, "density_matrix_pure_TLS: Failed to allocate dm (Out of memory?).\n");
                return NULL;
        }

	aa = sqrt(1-p_ex);
	bb = sqrt(p_ex);
	GSL_SET_COMPLEX(&z, aa * aa, 0.0); gsl_matrix_complex_set(dm, 0, 0, z);
	GSL_SET_COMPLEX(&z, aa * bb, 0.0); gsl_matrix_complex_set(dm, 0, 1, z);
	GSL_SET_COMPLEX(&z, bb * aa, 0.0); gsl_matrix_complex_set(dm, 1, 0, z);
	GSL_SET_COMPLEX(&z, bb * bb, 0.0); gsl_matrix_complex_set(dm, 1, 1, z);
	
	return dm;
}

/**
 * \breif	Initial density matrix for a 3-level systems.
 *
 * Generate a pure-state density matrix for a 3LS. (Used for constructing
 * initial states).
 *
 * p_ex = probability of being in the excited state
 *
 * @param	p_ex	Occupation probability of the two-level system excited state.
 */
gsl_matrix_complex *
density_matrix_pure_3LS(double p_ex)
{
	double aa, bb;
	gsl_matrix_complex *dm;
	gsl_complex z;
	
        if ((dm = gsl_matrix_complex_alloc(3,3)) == NULL)
        {
                fprintf(stderr, "density_matrix_pure_3LS: Failed to allocate dm (Out of memory?).\n");
                return NULL;
        }

	gsl_matrix_complex_set_zero(dm);
	
	GSL_SET_COMPLEX(&z, 1.0, 0.0);
	gsl_matrix_complex_set(dm, 0, 0, z);

	// XXX: fix this	

	return dm;
}	


/**
 * \breif	Initial density matrix for single-mode boson (thermal state)
 *
 * Generate a density matrix that represent a thermal boson state.
 * 
 * @param	w	Frequency (angular) of the boson field
 * @param	w_th	Temperature in frequency (angular) units.
 * @param	N	Number of fock states
 *
 */
gsl_matrix_complex *
density_matrix_boson_thermal(double w, double w_th, int N)
{
	gsl_matrix_complex *dm;
	gsl_complex z;
	double Z, p;
	int n;

	Z = exp(-w / (2*w_th)) / (1 - exp(-w/w_th));

    if ((dm = gsl_matrix_complex_alloc(N,N)) == NULL)
    {
        fprintf(stderr, "density_matrix_boson_thermal: Failed to allocate dm (Out of memory?).\n");
        return NULL;
	}
	
	gsl_matrix_complex_set_zero(dm);
	
	for (n = 0; n < N; n++)
	{
		p = exp(-w*(n + 0.5) / w_th) / Z;
		GSL_SET_COMPLEX(&z, p, 0.0);
		gsl_matrix_complex_set(dm, n, n, z);
	}
	
	return dm;
}


/**
 * \breif	Initial density matrix for a single-mode boson field (coherent state)
 *
 * Generate a density matrix that represents a coherent state:
 *
 * alpha = r exp(i theta)
 * @param       r        Coherent state amplitude.
 * @param       theta    Coherent state phase.
 * @param       N        Number of fock states.
 *
 */
gsl_matrix_complex *
density_matrix_coherent_state(double r, double theta, int N, int offset)
{
	int m, n;
	gsl_matrix_complex *dm;
	gsl_complex z;

	//printf("DEBUG: coherent state: alpha = %f exp(i*%f), truncated to %d states.\n", r, theta, N);
	
    if ((dm = gsl_matrix_complex_alloc(N,N)) == NULL)
    {
        fprintf(stderr, "density_matrix_coherent_state: Failed to allocate dm (Out of memory?).\n");
        return NULL;
    }

	for (m = 0; m < N; m++)
	{
		for (n = 0; n < N; n++)
		{	
			double dmij = exp(-r * r) * (pow(r, (double)(m+offset)) / fact_sqrt(m+offset)) * (pow(r, (double)(n+offset)) / fact_sqrt(n+offset));
			GSL_SET_COMPLEX(&z, dmij * cos((m-n) * theta), dmij * sin((m-n) * theta));
			gsl_matrix_complex_set(dm, m, n, z);
		}
	}

	return dm;
}

/**
 * \breif	coherent state wave-function
 *
 * Generate a wave matrix that represents a coherent state:
 *
 * alpha = r exp(i theta)
 * @param       r        Coherent state amplitude.
 * @param       theta    Coherent state phase.
 * @param       N        Number of fock states.
 *
 */
int 
wave_function_coherent_state(gsl_vector_complex *wf, double r, double theta, int N, int offset)
{
	int n;
	gsl_complex z;

    if (wf == NULL)
    {
        fprintf(stderr, "wave_function_coherent_state: Failed to allocate wf (Out of memory?).\n");
        return -1;
    }

        gsl_vector_complex_set_zero(wf);

        if (r == 0)
        {
                gsl_vector_complex_set(wf, 0, GSL_COMPLEX_ONE);
        }
        else
        {
	        for (n = 0; n < N; n++)
	        {	
		        double wfi = exp(-r * r / 2.0) * (pow(r, (double)(n+offset)) / fact_sqrt(n+offset));
		        GSL_SET_COMPLEX(&z, wfi * cos(n * theta), wfi * sin(n * theta));
		        gsl_vector_complex_set(wf, n, z);
	        }
        }

	return 0;
}


/**
 * \breif	Initial density matrix, Fock state.
 *
 * Generate a density matrix that represent a fock state
 *
 * @param	n	Fock state excitation number.
 * @param	N	Number of fock states in the system.
 *
 */
gsl_matrix_complex *
density_matrix_fock_state(int n, int N)
{
	gsl_matrix_complex *dm;
	
	if (n >= N)
	{
                fprintf(stderr, "density_matrix_fock_state: n >= N.\n");
                return NULL;
	}
	
        if ((dm = gsl_matrix_complex_alloc(N,N)) == NULL)
        {
                fprintf(stderr, "density_matrix_fock_state: Failed to allocate dm (Out of memory?).\n");
                return NULL;
        }

	gsl_matrix_complex_set_zero(dm);
	gsl_matrix_complex_set(dm, n, n, GSL_COMPLEX_ONE);
	
	return dm;
}

/**
 * \breif	Fock state wave function.
 *
 * Generate a wave-function that represent a fock state
 *
 * @param	n	Fock state excitation number.
 * @param	N	Number of fock states in the system.
 *
 */
gsl_vector_complex *
wave_function_fock_state(int n, int N)
{
	gsl_vector_complex *wf;
	
	if (n >= N)
	{
                fprintf(stderr, "wave_function_fock_state: n >= N.\n");
                return NULL;
	}
	
        if ((wf = gsl_vector_complex_alloc(N)) == NULL)
        {
                fprintf(stderr, "wave_function_fock_state: Failed to allocate wf (Out of memory?).\n");
                return NULL;
        }

	gsl_vector_complex_set_zero(wf);
	gsl_vector_complex_set(wf, n, GSL_COMPLEX_ONE);
	
	return wf;
}

/**
 * \breif	Calculate the expectation value
 *
 * Calculate the expectation value for the operator op for the system in the state
 * described by the density matrix rho.
 *
 * COULD BE OPTIMIZED.
 * 
 * @param	rho	Full system density matrix.
 * @param	op	Operator for which the expectation value is to be calculated.
 * 
 */
gsl_complex 
density_matrix_expectation_value(gsl_matrix_complex *rho, gsl_matrix_complex *op)
{
	gsl_matrix_complex *prod;
	gsl_complex alpha, beta, tr;
	int i;

	prod = gsl_matrix_complex_alloc(rho->size1, rho->size2);

	GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	GSL_SET_COMPLEX(&beta,  0.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho, op, beta, prod);

	//printf("rho =\n");
	//misc_matrix_complex_print(rho, rho->size1);
	//printf("op =\n");
	//misc_matrix_complex_print(op, rho->size1);
	//printf("prod =\n");
	//misc_matrix_complex_print(prod, rho->size1);
		 

	GSL_SET_COMPLEX(&tr, 0.0, 0.0);
	for (i = 0; i < rho->size1; i++)
		tr = gsl_complex_add(tr, gsl_matrix_complex_get(prod, i, i));


	gsl_matrix_complex_free(prod);
	
	//printf("trace = (%f,%f)\n", GSL_REAL(tr), GSL_IMAG(tr));
	return tr;
}

/**
 * \breif	Calculate the expectation value
 *
 * Calculate the expectation value for the operator op for the system in the state
 * described by the wave function wf.
 *
 * COULD BE OPTIMIZED.
 * 
 * @param	wf	Full system wave function
 * @param	op	Operator for which the expectation value is to be calculated.
 * 
 */
gsl_complex
wave_function_expectation_value(gsl_vector_complex *wf, gsl_matrix_complex *op)
{
	gsl_vector_complex *prod;
	gsl_complex expt;
	int i;

	prod = gsl_vector_complex_alloc(op->size1);
    gsl_blas_zgemv(CblasNoTrans, GSL_COMPLEX_ONE, op, wf, GSL_COMPLEX_ZERO, prod);
    gsl_blas_zdotu(wf, prod, &expt);

	gsl_vector_complex_free(prod);
	return expt;
}


gsl_complex
density_matrix_trace(gsl_matrix_complex *rho)
{
        int i;
	gsl_complex tr;

	GSL_SET_COMPLEX(&tr, 0.0, 0.0);
	for (i = 0; i < rho->size1; i++)
		tr = gsl_complex_add(tr, gsl_matrix_complex_get(rho, i, i));

	return tr;
}

/**
 * \breif	Calculate the concurrence of a two-level system.
 *
 * Calculate the concurrence for the density matrix (two TLS) rho.
 *
 * @param	qs	Data structure for specification of the quantum system
 * @param	rho	Density matrix of the system.
 *
 */
double
density_matrix_concurrence(quantum_system *qs, gsl_matrix_complex *rho)
{
	int ns;
	double c = 0.0;
	gsl_matrix_complex *y1, *y2, *y, *ws1, *ws2;
	gsl_vector_complex *eval = gsl_vector_complex_alloc(4);
        gsl_matrix_complex *evec = gsl_matrix_complex_alloc(4, 4);
      
        ns = quantum_system_nstates(qs);
	
	y1  = gsl_matrix_complex_alloc(ns, ns);
	y2  = gsl_matrix_complex_alloc(ns, ns);

	operator_sigma_y(y1, qs, 0);
	operator_sigma_y(y2, qs, 1); 

	y   = gsl_matrix_complex_alloc(ns, ns);
	ws1 = gsl_matrix_complex_alloc(ns, ns);
	ws2 = gsl_matrix_complex_alloc(ns, ns);


	// calculate: rho * (y1 * y2) * rho' * (y1 * y2)
	//            rho *     y     * rho' *     y
	//            ---------------   -----------------
	//                  ws1                ws2
	//                   
	gsl_blas_zgemm(CblasNoTrans,   CblasNoTrans, GSL_COMPLEX_ONE, y1, y2, GSL_COMPLEX_ZERO, y);
	
	gsl_blas_zgemm(CblasNoTrans,   CblasNoTrans, GSL_COMPLEX_ONE, rho, y, GSL_COMPLEX_ZERO, ws1);
	gsl_blas_zgemm(CblasConjTrans, CblasNoTrans, GSL_COMPLEX_ONE, rho, y, GSL_COMPLEX_ZERO, ws2);
	
	gsl_blas_zgemm(CblasNoTrans,   CblasNoTrans, GSL_COMPLEX_ONE, ws1, ws2, GSL_COMPLEX_ZERO, y);
	
	//printf("rr_real = \n"); misc_matrix_complex_print_real(y, 4);
	//printf("rr_imag = \n"); misc_matrix_complex_print_imag(y, 4);
	printf("rr = \n"); misc_matrix_complex_print(y, 4);

	// calculate eigenvalues
#ifndef RSCC	
	gsl_ext_eigen_zgeev(y, evec, eval);
#else
	fprintf(stderr, "WARNING: gsl_ext_eigen_zgeev note supported on RSCC.\n");
#endif
	{
		int i, j, n = 4;
		
        printf("Eigenvalues = \n");
        for (i = 0; i < n; i++)
        {
                gsl_complex z;
                z = gsl_vector_complex_get(eval, i);
                printf("\t(%f, %f)", GSL_REAL(z), GSL_IMAG(z));
        }
        printf("\n");

        printf("evec = \n");
        for (i = 0; i < n; i++)
        {
                for (j = 0; j < n; j++)
                {
                        gsl_complex z;
                        z = gsl_matrix_complex_get(evec, i, j);
                        printf("\t(%f, %f)", GSL_REAL(z), GSL_IMAG(z));
                }
                printf("\n");
        }
	}

	
	printf("c: eigvals: (%f, %f), (%f, %f), (%f, %f), (%f, %f)\n", 
	       GSL_REAL(gsl_vector_complex_get(eval, 0)), GSL_IMAG(gsl_vector_complex_get(eval, 0)), 
	       GSL_REAL(gsl_vector_complex_get(eval, 1)), GSL_IMAG(gsl_vector_complex_get(eval, 1)), 
	       GSL_REAL(gsl_vector_complex_get(eval, 2)), GSL_IMAG(gsl_vector_complex_get(eval, 2)), 
	       GSL_REAL(gsl_vector_complex_get(eval, 3)), GSL_IMAG(gsl_vector_complex_get(eval, 3)));

	//c = + sqrt(gsl_vector_get(eval, 0)) - sqrt(gsl_vector_get(eval, 1))
	//    - sqrt(gsl_vector_get(eval, 2)) - sqrt(gsl_vector_get(eval, 3));
	c = + GSL_REAL(gsl_vector_complex_get(eval, 0)) - GSL_REAL(gsl_vector_complex_get(eval, 1))
	    - GSL_REAL(gsl_vector_complex_get(eval, 2)) - GSL_REAL(gsl_vector_complex_get(eval, 3));

	c = (c < 0.0) ? 0.0 : c;

	gsl_matrix_complex_free(y1);
	gsl_matrix_complex_free(y2);
	gsl_matrix_complex_free(y);
	gsl_matrix_complex_free(ws1);
	gsl_matrix_complex_free(ws2);
	gsl_matrix_complex_free(evec);
	gsl_vector_complex_free(eval);
		
	return c;
}

/*------------------------------------------------------------------------------------------------
 * Density matrix and wave function operations:
 *
 */
int
density_matrix_mix(gsl_matrix_complex *rho_mix, gsl_matrix_complex *rho1, gsl_matrix_complex *rho2)
{
        gsl_complex z;
        GSL_SET_COMPLEX(&z, 0.5, 0.0);

        gsl_matrix_complex_memcpy(rho_mix, rho1);
        gsl_matrix_complex_add(rho_mix, rho2);
        gsl_matrix_complex_scale(rho_mix, z);

        return 0;
}

int
wave_function_superposition(gsl_vector_complex *wf_sp, gsl_vector_complex *wf1, gsl_vector_complex *wf2)
{
        gsl_complex z;
        GSL_SET_COMPLEX(&z, sqrt(0.5), 0.0);

        gsl_vector_complex_memcpy(wf_sp, wf1);
        gsl_vector_complex_add (wf_sp, wf2);
        gsl_vector_complex_scale(wf_sp, z);

        return 0;
}

int
wave_function_to_density_matrix(gsl_matrix_complex *rho, gsl_vector_complex *wf)
{
        gsl_complex x, y;
        int i, j;

        if ((rho->size1) != (wf->size) || (rho->size2) != (wf->size))
        {
                fprintf(stderr, "wave_function_to_density_matrix: vector-matrix size mismatch\n");
                return -1;
        }

        for (i = 0; i < wf->size; i++)
        {
                for (j = 0; j < wf->size; j++)
                {
                        x = gsl_vector_complex_get(wf, i);
                        y = gsl_vector_complex_get(wf, j);
                        gsl_matrix_complex_set(rho, i, j, gsl_complex_mul(x, gsl_complex_conjugate(y)));
                }
        }

        return 0;
}














