/* 
 * 
 * Robert Johansson <robert@riken.jp>
 *
 */

#include <stdio.h>
#include <math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

#include <gsl/gsl_eigen.h>
#include <gsl/gsl_integration.h>


#include "gsl_ext_expm_complex.h"
#include "distribution_functions.h"
#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "misc.h"
#include "gsl_ext.h"


static const size_t qawo_n=100;
static const size_t qawo_limit=1001;
static double abserr = 1e-6;
static double relerr = 1e-3;

/*
 * Coordinate to index and index to coordinate conversion functions:
 * Status: Done 
 */
static double
coordinate_from_index(double max, int i, int N)
{
    return (i - N/2.0) * max / (N/2.0);
}

static int
coordinate_to_index(double val, double max, int N)
{
    int n = (int)(floor((val + max) / (2*max) * (N)));

//    printf("debug: val+max = %f, 2*max = %f, (val + max) / (2*max) * N = %f, floor(...) = %f\n",
//            (val + max), (2*max), (val + max) / (2*max) * N, floor((val + max) / (2*max) * N));

    if (n < 0)
        n = 0;

    if (n >= N)
        n = N - 1;

//    printf("debug: coordinate_to_index: val = %f => n = %d: Val should be contained in [%f, %f]\n", 
//           val, n, coordinate_from_index(max, n, N), coordinate_from_index(max, n+1, N));

    return n;
}

/*
 * Interpolate an arbitrary point [between grid points] given a grid distribution.
 * Status: Done 
 */
static double
cf_wigner_interpolate(gsl_complex z, void *rawdata)
{
    wigner_data_t *data = (wigner_data_t *)rawdata;
      
    gsl_complex z00, z10, z01, f_z00, f_z01, f_z10, f_z;

    int z0_re_idx, z0_im_idx, z1_re_idx, z1_im_idx;

    double f_z_re, f_z_im;

//    printf("debug: cf_wigner_interpolate: lambda_max = %f + i %f\n", GSL_REAL(data->lambda_max), GSL_IMAG(data->lambda_max));
//    printf("debug: cf_wigner_interpolate: lambda = %f + i %f [%d]\n", GSL_REAL(lambda), GSL_IMAG(lambda), data->N);

    z0_re_idx = coordinate_to_index(GSL_REAL(z), GSL_REAL(data->lambda_max), data->N);
    z0_im_idx = coordinate_to_index(GSL_IMAG(z), GSL_IMAG(data->lambda_max), data->N);

    z0_re_idx = (z0_re_idx < 0) ? 0 : z0_re_idx;
    z0_im_idx = (z0_im_idx < 0) ? 0 : z0_im_idx;

    z1_re_idx = (z0_re_idx < (data->N-1)) ? z0_re_idx + 1 : z0_re_idx;
    z1_im_idx = (z0_im_idx < (data->N-1)) ? z0_im_idx + 1 : z0_im_idx;

 //   printf("debug: cf_wigner_interpolate idx_re(1,2) = (%d -> %d), idx_im(1,2) = (%d -> %d)\n", 
 //           z0_re_idx, z1_re_idx, z0_im_idx, z1_im_idx);
  
    GSL_SET_COMPLEX(&z00, 
        coordinate_from_index(GSL_REAL(data->lambda_max), z0_re_idx, data->N),
        coordinate_from_index(GSL_REAL(data->lambda_max), z0_im_idx, data->N)); 

    GSL_SET_COMPLEX(&z01, 
        coordinate_from_index(GSL_REAL(data->lambda_max), z0_re_idx, data->N),
        coordinate_from_index(GSL_REAL(data->lambda_max), z1_im_idx, data->N)); 

    GSL_SET_COMPLEX(&z10, 
        coordinate_from_index(GSL_REAL(data->lambda_max), z1_re_idx, data->N),
        coordinate_from_index(GSL_REAL(data->lambda_max), z0_im_idx, data->N)); 

    f_z00 = gsl_matrix_complex_get(data->Cw, z0_re_idx, z0_im_idx);
    f_z01 = gsl_matrix_complex_get(data->Cw, z0_re_idx, z1_im_idx);
    f_z10 = gsl_matrix_complex_get(data->Cw, z1_re_idx, z0_im_idx);

    printf("debug: cf_wigner_interpolate: z           = %f + i %f. f(z) = ?\n", 
           GSL_REAL(z), GSL_IMAG(z));
    printf("debug: cf_wigner_interpolate: z00[%d, %d] = %f + i %f. f(z00) = %f + i %f\n", 
           z0_re_idx, z0_im_idx, GSL_REAL(z00), GSL_IMAG(z00), GSL_REAL(f_z00), GSL_IMAG(f_z00));
    printf("debug: cf_wigner_interpolate: z01[%d, %d] = %f + i %f. f(z01) = %f + i %f\n", 
           z0_re_idx, z1_im_idx, GSL_REAL(z01), GSL_IMAG(z01), GSL_REAL(f_z01), GSL_IMAG(f_z01));
    printf("debug: cf_wigner_interpolate: z10[%d, %d] = %f + i %f. f(z10) = %f + i %f\n", 
           z1_re_idx, z0_im_idx, GSL_REAL(z10), GSL_IMAG(z10), GSL_REAL(f_z10), GSL_IMAG(f_z10));


    printf("debug: cf_wigner_interpolate: dx = %f [%f], dy = %f [%f]\n", 
           GSL_REAL(z)   - GSL_REAL(z00),
           GSL_REAL(z10) - GSL_REAL(z00),
           GSL_IMAG(z)   - GSL_IMAG(z00),
           GSL_IMAG(z01) - GSL_IMAG(z00));

    f_z_re = GSL_REAL(f_z00);
    if ((z1_re_idx != z0_re_idx) && GSL_REAL(z10) != GSL_REAL(z00))
    {
        f_z_re += (GSL_REAL(f_z10) - GSL_REAL(f_z00)) * 
                  (GSL_REAL(z) - GSL_REAL(z00)) / (GSL_REAL(z10) - GSL_REAL(z00));
    }
    if ((z1_im_idx != z0_im_idx) && (GSL_IMAG(z01) != GSL_IMAG(z00)))
    {
        f_z_re += (GSL_REAL(f_z01) - GSL_REAL(f_z00)) * 
                  (GSL_IMAG(z) - GSL_IMAG(z00)) / (GSL_IMAG(z01) - GSL_IMAG(z00));
    }

/*
    f_z_im = GSL_IMAG(f_z00);
    if(0) //if (GSL_REAL(z10) != GSL_REAL(z00))
    {
        f_z_im += (GSL_IMAG(f_z10) - GSL_IMAG(f_z00)) * 
                  (GSL_REAL(z) - GSL_REAL(z00)) / (GSL_REAL(z10) - GSL_REAL(z00));
    }
    if(0) //if (GSL_IMAG(z01) != GSL_IMAG(z00))
    {
        f_z_im += (GSL_IMAG(f_z01) - GSL_IMAG(f_z00)) * 
                  (GSL_IMAG(z) - GSL_IMAG(z00)) / (GSL_IMAG(z01) - GSL_IMAG(z00));
    }
*/
    if (isnan(f_z_re))
    {
        printf("cf_wigner_interpolate: z_re was NaN... \n");
        f_z_re = 0.0;
    }

/*
    if (isnan(f_z_im))
    {
        printf("cf_wigner_interpolate: z_im was NaN... \n");
        f_z_im = 0.0;
    }
*/

//    GSL_SET_COMPLEX(&f_z, f_z_re, 0.0);
//    GSL_SET_COMPLEX(&f_z, exp(-gsl_complex_abs(z)), 0.0);
//    GSL_SET_COMPLEX(&f_z, 0.0, 0.0);

//    return f_z;

   // f_z_re = exp(-gsl_complex_abs(z));

    return f_z_re;
//    return ;
}


/**
 * \breif	Calculate the Q-function corresponding to a system with density matrix rho.
 *
 */
int
distribution_function_Q(gsl_matrix_complex *Q, gsl_matrix_complex *RHO, double alpha_max)
{
	gsl_vector_complex *wf, *prod;
    gsl_complex z;
    int Qsize, RHOsize, i, j;

    double a_re, a_im, r, theta;

    if (Q == NULL || RHO == NULL)
    {
        printf("distribution_function_Q: parameter errors\n");
        return -1;
    }

    Qsize   =   Q->size1;
    RHOsize = RHO->size1;

    wf   = gsl_vector_complex_alloc(RHOsize);     
    prod = gsl_vector_complex_alloc(RHOsize);     
	gsl_matrix_complex_set_zero(Q);
        
        for (i = 0; i < Qsize; i++)
        {
                a_re = coordinate_from_index(alpha_max, i, Qsize);
                for (j = 0; j < Qsize; j++)
                {
                        a_im = coordinate_from_index(alpha_max, j, Qsize);
                        
                        r     = sqrt(a_re * a_re + a_im * a_im);
                        theta = atan(a_im / a_re);
                        if (a_re < 0)
                        {
                                theta = theta + M_PI;
                        }

                        wave_function_coherent_state(wf, r, theta, RHOsize, 0);

                        gsl_blas_zgemv(CblasNoTrans, GSL_COMPLEX_ONE, RHO, wf, GSL_COMPLEX_ZERO, prod);

                        gsl_blas_zdotc(wf, prod, &z);

                        gsl_matrix_complex_set(Q, i, j, z);
                }
        }
        
        return 0;       
}


/*
 * Compute the characteristic function corresponding to an harmonic
 * oscillator density matrix:
 */
int
distribution_function_characteristic_w(
            gsl_matrix_complex *phase_space, 
            gsl_matrix_complex *a,
            gsl_matrix_complex *ad,
            gsl_matrix_complex *rho,
            gsl_complex lambda_max)
{
    static char *func_name = "distribution_function_characteristic_w";

    gsl_matrix_complex *a_exp, *ad_exp, *ws;

    int i, j, qsn, N;

    gsl_complex lambda, z;

    if (!phase_space || !a || !ad || !rho)
    {
        fprintf(stderr, "%s: invalid arguments (null pointer).\n", func_name);
        return 0;
    }
    else
    {
        printf("%s: calculating... \n", func_name);
    }
    
    N   = phase_space->size1;
    qsn = rho->size1;

    if ( (rho->size2 != qsn) ||
         (a->size1   != qsn) ||
         (a->size2   != qsn) ||
         (ad->size1  != qsn) ||
         (ad->size2  != qsn) )
    {
        fprintf(stderr, "%s: invalid arguments (size mismatch).\n", func_name);
        return 0;
    }

    ws     = gsl_matrix_complex_alloc(qsn, qsn);
    a_exp  = gsl_matrix_complex_alloc(qsn, qsn);
    ad_exp = gsl_matrix_complex_alloc(qsn, qsn);

    // i - real
    // j - imag
    for (i = 0; i < N; i++)
    {
        for (j = 0; j < N; j++)
        {
            // calculate the exponents of a and ad:
            GSL_SET_COMPLEX(&lambda, 
                coordinate_from_index(GSL_REAL(lambda_max), i, N),
                coordinate_from_index(GSL_IMAG(lambda_max), j, N));

            gsl_matrix_complex_memcpy(ws, a);
            gsl_matrix_complex_scale(ws, gsl_complex_negative(gsl_complex_conjugate(lambda)));
            gsl_linalg_complex_exponential_ss(ws, a_exp, GSL_PREC_DOUBLE);

            gsl_matrix_complex_memcpy(ws, ad);
            gsl_matrix_complex_scale(ws, lambda);
            gsl_linalg_complex_exponential_ss(ws, ad_exp, GSL_PREC_DOUBLE);

            gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, ad_exp, a_exp, GSL_COMPLEX_ZERO, ws);

            z = density_matrix_expectation_value(rho, ws);

            gsl_matrix_complex_set(phase_space, i, j, gsl_complex_mul_real(z, exp(-0.5 * gsl_complex_abs2(lambda))));
        }
    }    

    gsl_matrix_complex_free(ws);
    gsl_matrix_complex_free(a_exp);
    gsl_matrix_complex_free(ad_exp);

    return 0;
}


double
cb_integrand_wigner_cf(gsl_complex lambda, gsl_matrix_complex *phase_space_cf, gsl_complex alpha)
{
    
}

static double
cb_qawo_wigner_f_im(double lambda_im, void *rawdata)
{
//    gsl_complex z;
    double z;
    wigner_data_t *data = (wigner_data_t *)rawdata;

    // set the real and imaginary part
    GSL_SET_COMPLEX(&(data->lambda), GSL_REAL(data->lambda), lambda_im);

    z = cf_wigner_interpolate(data->lambda, data);

//    printf("debug: cb_wigner_f_im [lambda = %f + i %f]: z = %f + i %f\n", 
//           GSL_REAL(data->lambda), GSL_IMAG(data->lambda), z, 0.0); //GSL_REAL(z), GSL_IMAG(z));

//    return GSL_REAL(z);
    return z;
}

static double
cb_qawo_wigner_f_re(double lambda_re, void *rawdata)
{
    double result = 1.0, abserr;
    int status;

    wigner_data_t *data = (wigner_data_t *)rawdata;

//    printf("debug [f_re(%f)]: calculating...\n", lambda_re);

    // set the real part
    GSL_SET_COMPLEX(&(data->lambda), lambda_re, 0);

//    data->wf_im = gsl_integration_qawo_table_alloc(
//                      2 * GSL_REAL(data->alpha), 
//                      2 * GSL_IMAG(data->lambda_max), 
//                      GSL_INTEG_COSINE, qawo_n); 
//    gsl_integration_qawo_table_set_length(data->wf_im, 2 * GSL_IMAG(data->lambda_max));

    gsl_integration_qawo_table_set(data->wf_im, 2 * GSL_REAL(data->alpha), 2 * GSL_IMAG(data->lambda_max), GSL_INTEG_COSINE);


    if ((status = gsl_integration_qawo(&(data->f_im), -GSL_IMAG(data->lambda_max), abserr, relerr, qawo_limit, data->w_im, data->wf_im, &result, &abserr)) != 0)
    {
        printf("Warning: gsl_integration_qawo failed [status = %d]...\n", status);
    }
    else
    {
    //    printf("debug [f_re(%f)]: gsl_integration_qawo converged to %f with abserr %f\n", lambda_re, result, abserr);
    }        

//    gsl_integration_qawo_table_free(data->wf_im);

    return result;
}


/*
 * Compute the Wigner function corresponding to an harmonic
 * oscillator density matrix, by 2d numberical integration
 * of the characteristic function.
 */
int
distribution_function_wigner_cf_qawo(
            gsl_matrix_complex *phase_space_wigner, 
            gsl_matrix_complex *a,
            gsl_matrix_complex *ad,
            gsl_matrix_complex *rho,
            gsl_complex alpha_max)
{
    static char *func_name = "distribution_function_wigner_cf_qawo";

    gsl_matrix_complex *phase_space_cfw;
    int i, j, qsn, N, status;
    double result, abserr;
    gsl_complex z;

    wigner_data_t data;

    N   = phase_space_wigner->size1;
    data.N = N;
    data.alpha_max  = alpha_max;
    data.lambda_max = alpha_max;

//    phase_space_cfw = gsl_matrix_complex_alloc(phase_space_wigner->size1, phase_space_wigner->size2);
    data.Cw = gsl_matrix_complex_alloc(phase_space_wigner->size1, phase_space_wigner->size2);

    // first calculate the characteristic function
    // lambda_max = 2*pi * alpha_max ???
    distribution_function_characteristic_w(data.Cw, a, ad, rho, data.lambda_max); 
    
  
//    data.Cw = phase_space_cfw;

    // set-up integration spaces
    data.w_re  = gsl_integration_workspace_alloc(qawo_limit);
    data.w_im  = gsl_integration_workspace_alloc(qawo_limit);

    data.f_re.function = &cb_qawo_wigner_f_re;
    data.f_re.params   = &data;
    data.f_im.function = &cb_qawo_wigner_f_im; 
    data.f_im.params   = &data;

    data.wf_im = gsl_integration_qawo_table_alloc(2 * GSL_IMAG(data.alpha), 2 * GSL_IMAG(data.lambda_max), GSL_INTEG_COSINE, qawo_n); 
    gsl_integration_qawo_table_set_length(data.wf_im, 2 * GSL_IMAG(data.lambda_max));
    data.wf_re = gsl_integration_qawo_table_alloc(2 * GSL_IMAG(data.alpha), 2 * GSL_REAL(data.lambda_max), GSL_INTEG_COSINE, qawo_n); 
    gsl_integration_qawo_table_set_length(data.wf_re, 2 * GSL_REAL(data.lambda_max));

    // i - real
    // j - imag
    for (i = 0; i < N; i++)
    {
        for (j = 0; j < N; j++)
        {
            GSL_SET_COMPLEX(&(data.alpha), 
                coordinate_from_index(GSL_REAL(alpha_max), i, N),
                coordinate_from_index(GSL_IMAG(alpha_max), j, N));
       
//            data.wf_re = gsl_integration_qawo_table_alloc(
//                             2 * GSL_IMAG(alpha), 
//                             2 * GSL_REAL(lambda_max), 
//                             GSL_INTEG_COSINE, qawo_n); 

              gsl_integration_qawo_table_set(data.wf_re, 2 * GSL_IMAG(data.alpha), 2 * GSL_REAL(data.lambda_max), GSL_INTEG_COSINE);
//            gsl_integration_qawo_table_set_length(data.wf_re, 2 * GSL_REAL(lambda_max));

            if ((status = gsl_integration_qawo(&(data.f_re), -GSL_REAL(data.lambda_max), abserr, relerr, qawo_limit, data.w_re, data.wf_re, &result, &abserr)) != 0)
            {
                printf("Warning: gsl_integration_qawo failed [status = %d]...\n", status);
            }
            else
            {
                printf("debug: alpha = [%f, %f]: gsl_integration_qawo converged to %f with abserr %f\n", 
                        GSL_REAL(data.alpha), GSL_IMAG(data.alpha), result, abserr);
            }
            
            //result = cf_wigner_interpolate(data.alpha, (void *)&data);

            GSL_SET_COMPLEX(&z, result, 0.0);
            //z = gsl_matrix_complex_get(data.Cw, i, j);

            gsl_matrix_complex_set(phase_space_wigner, i, j, z);
            printf("debug: W(alpha = [%f, %f]) = [%f, %f]\n", 
                   GSL_REAL(data.alpha), GSL_IMAG(data.alpha), GSL_REAL(z), GSL_IMAG(z));

//            gsl_integration_qawo_table_free(data.wf_re);
        }
    }    

    gsl_integration_workspace_free(data.w_re);
    gsl_integration_workspace_free(data.w_im);

    return 0;
}

/*
 * Compute the Wigner function corresponding to an harmonic
 * oscillator density matrix, by 2d FFT of the characteristic function.
 */
int
distribution_function_wigner_cf_fft(
            gsl_matrix_complex *phase_space_wigner, 
            gsl_matrix_complex *a,
            gsl_matrix_complex *ad,
            gsl_matrix_complex *rho,
            gsl_complex alpha_max)
{
    static char *func_name = "distribution_function_wigner_cf_quad";

    gsl_matrix_complex *phase_space_cf;
    int i, j, qsn, N;
    gsl_complex alpha, lambda_max, z;

    phase_space_cf = gsl_matrix_complex_alloc(phase_space_wigner->size1, phase_space_wigner->size2);

    // first calculate the characteristic function
    // lambda_max = 2*pi * alpha_max ???
    distribution_function_characteristic_w(phase_space_cf, a, ad, rho, lambda_max); 
    
    //gsl_ifft(phase_space_wigner, phase_space_cf);

    return 0;
}

/*
 * Compute the Wigner function corresponding to an harmonic
 * oscillator density matrix, by 2d numberical integration
 * of the characteristic function.
 */
static double
cb_qags_wigner_f_im(double lambda_im, void *rawdata)
{
    double z;
    wigner_data_t *data = (wigner_data_t *)rawdata;

    GSL_SET_COMPLEX(&(data->lambda), GSL_REAL(data->lambda), lambda_im);
    //z = exp(-gsl_complex_abs2(data->alpha)*gsl_complex_abs2(data->lambda));
    z = cf_wigner_interpolate(data->lambda, data);
    return z;
}

static double
cb_qags_wigner_f_re(double lambda_re, void *rawdata)
{
    double result, error;
    int status;

    wigner_data_t *data = (wigner_data_t *)rawdata;

    GSL_SET_COMPLEX(&(data->lambda), lambda_re, 0);

    if ((status = gsl_integration_qags(
                    &(data->f_im), 
                    -GSL_IMAG(data->lambda_max),
                    +GSL_IMAG(data->lambda_max), 
                    abserr, relerr, 
                    10000, data->w_im, 
                    &result, &error)) != 0)
    {
        printf("Warning: gsl_integration_qawo failed [status = %d]...\n", status);
    }
    else
    {
     //  printf("debug [f_re(%f)]: gsl_integration_qawo converged to %f with abserr %f [interval = %d]\n", lambda_re, result, error, (int)(data->w_im->size));
    }        

    return result;
//    return exp(-gsl_complex_abs2(data->alpha)*gsl_complex_abs2(data->lambda));
}

int
distribution_function_wigner_cf_quad(
            gsl_matrix_complex *phase_space_wigner, 
            gsl_matrix_complex *a,
            gsl_matrix_complex *ad,
            gsl_matrix_complex *rho,
            gsl_complex alpha_max)
{
    static char *func_name = "distribution_function_wigner_cf_quad";

    gsl_matrix_complex *phase_space_cfw;
    int i, j, status;
    double result, error;
    gsl_complex z;

    wigner_data_t data;

    data.N = phase_space_wigner->size1;
    data.alpha_max  = alpha_max;
    data.lambda_max = alpha_max;

    // first calculate the characteristic function
    data.Cw = gsl_matrix_complex_alloc(phase_space_wigner->size1, phase_space_wigner->size2);
    distribution_function_characteristic_w(data.Cw, a, ad, rho, data.lambda_max); 

    // set-up integration spaces
    data.w_re = gsl_integration_workspace_alloc(10000);
    data.w_im = gsl_integration_workspace_alloc(10000);
       
    data.f_re.function = &cb_qags_wigner_f_re;
    data.f_re.params   = &data;
    data.f_im.function = &cb_qags_wigner_f_im; 
    data.f_im.params   = &data;

    // i - real
    // j - imag
    for (i = 0; i < data.N; i++)
    {
        for (j = 0; j < data.N; j++)
        {
            GSL_SET_COMPLEX(&(data.alpha), 
                coordinate_from_index(GSL_REAL(alpha_max), i, data.N),
                coordinate_from_index(GSL_IMAG(alpha_max), j, data.N));
       
            if ((status = gsl_integration_qags(
                                &(data.f_re), 
                                -GSL_REAL(data.lambda_max),
                                +GSL_REAL(data.lambda_max), 
                                abserr, relerr, 
                                10000, data.w_re, 
                                &result, &error)) != 0)
            {
                printf("Warning: gsl_integration_qawo failed [status = %d]...\n", status);
            }
            else
            {
                printf("debug: alpha = [%f, %f]: gsl_integration_qawo converged to %f with abserr %f [interval = %d]\n", 
                        GSL_REAL(data.alpha), GSL_IMAG(data.alpha), result, error, (int)(data.w_re->size));
            }
            
            GSL_SET_COMPLEX(&z, result, 0.0);
            gsl_matrix_complex_set(phase_space_wigner, i, j, z);
        }
    }    

    gsl_integration_workspace_free(data.w_re);
    gsl_integration_workspace_free(data.w_im);

    return 0;
}




/*
 * Integrand for quadrature:
 * XXX: unfinished
 */
double
cb_integrand_wigner_hermite(double x, void *data)
{
    wigner_data_t *integrand_data =  (wigner_data_t *)data;
    
//    double x = 0;
    int n, N = 0;

    for (n = 0; n < N; n++)
    {
//        x += GSL_REAL(gsl_matrix_complex_get(integrand_data->rho, n, n)) *
//             hermite_poly_eval(n, x + integrand_data->q / 2) *
//             hermite_poly_eval(n, x - integrand_data->q / 2);
    }

    return x;
}

/*
 * XXX unfinished
 */
int
distribution_function_wigner_hermite_quad(
            gsl_matrix_complex *phase_space_wigner, 
            gsl_matrix_complex *rho,
            double x_limit)
{
    static char *func_name = "distribution_function_wigner_cf_quad";

    gsl_matrix_complex *phase_space_cf;
    int i, j, qsn, N, M;
    gsl_complex alpha, lambda_max, z;
    double q, p; 

    wigner_data_t integrand_data;
   
    N = phase_space_wigner->size1;
    M = phase_space_wigner->size2;

    // i - real
    // j - imag
    for (i = 0; i < N; i++)
    {
        for (j = 0; j < M; j++)
        {
            integrand_data.q = (i - N/2) * x_limit / (N/2);
            integrand_data.p = (j - M/2) * x_limit / (M/2);
            
    //        z = gsl_quad_1d(-x_limit, x_limit,
    //                        cb_integrand_wigner_hermite,
    //                        (void *)&integrand_data);

            gsl_matrix_complex_set(phase_space_wigner, i, j, z);
        }
    }    

    return 0;
}



int
distribution_function_wigner_cf_sum(
            gsl_matrix_complex *phase_space_wigner, 
            gsl_matrix_complex *a,
            gsl_matrix_complex *ad,
            gsl_matrix_complex *rho,
            gsl_complex alpha_max)
{
    static char *func_name = "distribution_function_wigner_cf_quad";

    gsl_matrix_complex *phase_space_cfw;
    int i, j, m, n, status;
    double result, result_inner, error, dlambda_im, dlambda_re;
    gsl_complex z, ez, sum_outer, sum_inner;

    wigner_data_t data;

    data.N = phase_space_wigner->size1;
    data.alpha_max  = alpha_max;
    data.lambda_max = alpha_max;

    printf("rho = [%d, %d], a = [%d, %d], ad = [%d, %d]\n",
            rho->size1, rho->size2,
            a->size1, a->size2,
            ad->size1, ad->size2);


    // first calculate the characteristic function
    data.Cw = gsl_matrix_complex_alloc(phase_space_wigner->size1, phase_space_wigner->size2);
    distribution_function_characteristic_w(data.Cw, a, ad, rho, data.lambda_max); 

    dlambda_re = 2 * GSL_REAL(data.lambda_max) / data.N;
    dlambda_im = 2 * GSL_IMAG(data.lambda_max) / data.N;

    // i - real
    // j - imag
    for (i = 0; i < data.N; i++)
    {
        for (j = 0; j < data.N; j++)
        {
            GSL_SET_COMPLEX(&(data.alpha), 
                coordinate_from_index(GSL_REAL(alpha_max), i, data.N),
                coordinate_from_index(GSL_IMAG(alpha_max), j, data.N));

            //result = 0.0;
            GSL_SET_COMPLEX(&sum_outer, 0.0, 0.0);

            for (m = 0; m < data.N; m++)
            {
                //result_inner = 0.0;
                GSL_SET_COMPLEX(&sum_inner, 0.0, 0.0);

                for (n = 0; n < data.N; n++)
                {
                    GSL_SET_COMPLEX(&(data.lambda), 
                        coordinate_from_index(GSL_REAL(data.lambda_max), m, data.N),
                        coordinate_from_index(GSL_IMAG(data.lambda_max), n, data.N));

                    // value of the characteristic function
                    z = gsl_matrix_complex_get(data.Cw, m, n);

//                    result_inner += cos(2*GSL_IMAG(data.lambda)*GSL_REAL(data.alpha)) * GSL_REAL(z) * dlambda_im;

                    ez = gsl_complex_mul_real(
                            gsl_complex_exp(
                               gsl_complex_sub(
                                    gsl_complex_mul(gsl_complex_conjugate(data.lambda), data.alpha),
                                    gsl_complex_mul(data.lambda, gsl_complex_conjugate(data.alpha))
                                              )
                                           ), dlambda_im);

//                    ez = gsl_complex_exp(la);
//                    GSL_SET_COMPLEX(&ez, cos(GSL_REAL(la)), sin(GSL_IMAG(la));
                   
                 
                    sum_inner = gsl_complex_add(sum_inner, gsl_complex_mul(ez, z));                

                    // including sin * sin
//                    result_inner += (cos(2*GSL_REAL(data.lambda)*GSL_IMAG(data.alpha)) * cos(2*GSL_IMAG(data.lambda)*GSL_REAL(data.alpha)) + 
  //                                   sin(2*GSL_REAL(data.lambda)*GSL_IMAG(data.alpha)) * sin(2*GSL_IMAG(data.lambda)*GSL_REAL(data.alpha))
    //                                ) * GSL_REAL(z) * dlambda_im;
//

                }
//                result += cos(2*GSL_REAL(data.lambda)*GSL_IMAG(data.alpha)) * result_inner  * dlambda_re;
//                sum_inner = gsl_complex_mul_real(result_inner, dlambda_re);

                sum_outer = gsl_complex_add(sum_outer, gsl_complex_mul_real(sum_inner, dlambda_re));
//                result += result_inner  * dlambda_re;
            }
            //result = result / (M_PI * M_PI);
            sum_outer = gsl_complex_mul_real(sum_outer, 1.0/(M_PI * M_PI));

//            GSL_SET_COMPLEX(&z, result, 0.0);
            gsl_matrix_complex_set(phase_space_wigner, i, j, sum_outer);
        }
    }    

    return 0;
}


