/*
 * Functions for calculation of entanglement. 
 * 
 * 
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: entanglement.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */

#include <stdio.h>
#include <math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

#include <gsl/gsl_eigen.h>


#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "misc.h"
#include "gsl_ext.h"

/**
 * \breif	Calculate the von Neumann Entropy of a system with density matrix rho.
 *
 * This function takes a density matrix as input and calculates the von Neumann entropy.
 * Note that the density matrix is not restricted to be representing any particular
 * for of quantum system.
 *
 * @param	rho	The density matrix for which the von Neumann entropy is to be calculated.
 *
 */
double
entanglement_neumann_entropy(gsl_matrix_complex *rho)
{
	int i;
	double e, ne = 0.0;
	gsl_matrix_complex *rho_w;
	gsl_eigen_herm_workspace *w;
	gsl_vector *eval;

	rho_w = gsl_matrix_complex_alloc(rho->size1,rho->size1);
	gsl_matrix_complex_memcpy(rho_w, rho);
		
	eval = gsl_vector_alloc(rho->size1);
	w = gsl_eigen_herm_alloc(rho->size1);
	gsl_eigen_herm(rho_w, eval, w);

	for (i = 0; i < rho->size1; i++)
	{
		e = gsl_vector_get(eval, i);
		//ne += - e * log2(e);
		ne += - e * log(e);
	}

	/* clean-up */
	gsl_matrix_complex_free(rho_w);
	gsl_eigen_herm_free(w);	
	gsl_vector_free(eval);

	return ne;
}

/**
 * \breif	Calculate the log negativity for the density matrix of two TLS, rho.
 *
 * Calculate the negativity of a two-qubit system. 
 *
 * @param	rho	Density matrix for a two-qubit system.
 *
 */
double
entanglement_log_neg(gsl_matrix_complex *rho)
{
	int i;
	double ln, sum_eval;
	gsl_matrix_complex *rho_pt;
	gsl_eigen_herm_workspace *w;
	gsl_vector *eval;

	/* -- calculate partial transpose of rho, and store in rho_pt 
	 *
	 * | x x A B |      | x x a b |
	 * | x x C D |  ->  | x x c d |
	 * | a b y y |      | A B y y |
	 * | c d y y |      | C D y y |
	 *
	 */
	rho_pt = gsl_matrix_complex_alloc(4,4);
	gsl_matrix_complex_memcpy(rho_pt, rho);
	// A,B,C,D -> a,b,c,d
	gsl_matrix_complex_set(rho_pt, 0, 2, gsl_matrix_complex_get(rho, 2, 0));
	gsl_matrix_complex_set(rho_pt, 0, 3, gsl_matrix_complex_get(rho, 2, 1));
	gsl_matrix_complex_set(rho_pt, 1, 2, gsl_matrix_complex_get(rho, 3, 0));
	gsl_matrix_complex_set(rho_pt, 1, 3, gsl_matrix_complex_get(rho, 3, 1));
	// a,b,c,d -> A,B,C,D
	gsl_matrix_complex_set(rho_pt, 2, 0, gsl_matrix_complex_get(rho, 0, 2));
	gsl_matrix_complex_set(rho_pt, 2, 1, gsl_matrix_complex_get(rho, 0, 3));
	gsl_matrix_complex_set(rho_pt, 3, 0, gsl_matrix_complex_get(rho, 1, 2));
	gsl_matrix_complex_set(rho_pt, 3, 1, gsl_matrix_complex_get(rho, 1, 3));

	/* --- diagonalize rho_pt --- */
        eval = gsl_vector_alloc(4);
        w = gsl_eigen_herm_alloc(4);
	gsl_eigen_herm(rho_pt, eval, w);
        
//  	  printf("evals of rho_pt =\n");
//        for (i = 0; i < 4; i++)
//        {
//                double eval_i = gsl_vector_get(eval, i);
//                printf("\t%f", eval_i);
//        }
//        printf("\n");
	
	/* --- calculate the log negativity --- */
	sum_eval = 0;
	for (i = 0; i < 4; i++)
	{
		sum_eval += fabs(gsl_vector_get(eval, i));
	}
	ln = log2(sum_eval);

	/* clean-up */
	gsl_matrix_complex_free(rho_pt);
	gsl_eigen_herm_free(w);	
	gsl_vector_free(eval);

	return ln;
}


/**
 * \breif	Calculate the concurrence for the density matrix (two TLS) rho.
 *
 * Calculate the concurrence for the density matrix of a two-qubit system.
 *
 * @param	qs	Data structure that specifies the form of the quantum system.
 * @param	rho	The density matrix for a two-qubit system.
 *
 */
double
entanglement_concurrence(quantum_system *qs, gsl_matrix_complex *rho)
{
	int ns;
	double c = 0.0;
	gsl_matrix_complex *y1, *y2, *y, *ws1, *ws2;
	gsl_vector_complex *eval = gsl_vector_complex_alloc(4);
        gsl_matrix_complex *evec = gsl_matrix_complex_alloc(4, 4);
      
        ns = quantum_system_nstates(qs);

	y1  = gsl_matrix_complex_alloc(ns, ns);
	y2  = gsl_matrix_complex_alloc(ns, ns);
	
	operator_sigma_y(y1, qs, 0);
	operator_sigma_y(y2, qs, 1); 

	y   = gsl_matrix_complex_alloc(ns, ns);
	ws1 = gsl_matrix_complex_alloc(ns, ns);
	ws2 = gsl_matrix_complex_alloc(ns, ns);


	// calculate: rho * (y1 * y2) * rho' * (y1 * y2)
	//            rho *     y     * rho' *     y
	//            ---------------   -----------------
	//                  ws1                ws2
	//                   
	gsl_blas_zgemm(CblasNoTrans,   CblasNoTrans, GSL_COMPLEX_ONE, y1, y2, GSL_COMPLEX_ZERO, y);
	
	gsl_blas_zgemm(CblasNoTrans,   CblasNoTrans, GSL_COMPLEX_ONE, rho, y, GSL_COMPLEX_ZERO, ws1);
	gsl_blas_zgemm(CblasConjTrans, CblasNoTrans, GSL_COMPLEX_ONE, rho, y, GSL_COMPLEX_ZERO, ws2);
	
	gsl_blas_zgemm(CblasNoTrans,   CblasNoTrans, GSL_COMPLEX_ONE, ws1, ws2, GSL_COMPLEX_ZERO, y);
	
	//printf("rr_real = \n"); misc_matrix_complex_print_real(y, 4);
	//printf("rr_imag = \n"); misc_matrix_complex_print_imag(y, 4);
	printf("rr = \n"); misc_matrix_complex_print(y, 4);

	// calculate eigenvalues
	gsl_ext_eigen_zgeev(y, evec, eval);

	{
		int i, j, n = 4;
		
        printf("Eigenvalues = \n");
        for (i = 0; i < n; i++)
        {
                gsl_complex z;
                z = gsl_vector_complex_get(eval, i);
                printf("\t(%f, %f)", GSL_REAL(z), GSL_IMAG(z));
        }
        printf("\n");

        printf("evec = \n");
        for (i = 0; i < n; i++)
        {
                for (j = 0; j < n; j++)
                {
                        gsl_complex z;
                        z = gsl_matrix_complex_get(evec, i, j);
                        printf("\t(%f, %f)", GSL_REAL(z), GSL_IMAG(z));
                }
                printf("\n");
        }
	}

	
	printf("c: eigvals: (%f, %f), (%f, %f), (%f, %f), (%f, %f)\n", 
	       GSL_REAL(gsl_vector_complex_get(eval, 0)), GSL_IMAG(gsl_vector_complex_get(eval, 0)), 
	       GSL_REAL(gsl_vector_complex_get(eval, 1)), GSL_IMAG(gsl_vector_complex_get(eval, 1)), 
	       GSL_REAL(gsl_vector_complex_get(eval, 2)), GSL_IMAG(gsl_vector_complex_get(eval, 2)), 
	       GSL_REAL(gsl_vector_complex_get(eval, 3)), GSL_IMAG(gsl_vector_complex_get(eval, 3)));

	//c = + sqrt(gsl_vector_get(eval, 0)) - sqrt(gsl_vector_get(eval, 1))
	//    - sqrt(gsl_vector_get(eval, 2)) - sqrt(gsl_vector_get(eval, 3));
	c = + GSL_REAL(gsl_vector_complex_get(eval, 0)) - GSL_REAL(gsl_vector_complex_get(eval, 1))
	    - GSL_REAL(gsl_vector_complex_get(eval, 2)) - GSL_REAL(gsl_vector_complex_get(eval, 3));

	c = (c < 0.0) ? 0.0 : c;

	gsl_matrix_complex_free(y1);
	gsl_matrix_complex_free(y2);
	gsl_matrix_complex_free(y);
	gsl_matrix_complex_free(ws1);
	gsl_matrix_complex_free(ws2);
	gsl_matrix_complex_free(evec);
	gsl_vector_complex_free(eval);
		
	return c;
}



