/**
 * @mainpage Hamiltonians
 *
 * Setup Hamiltonians for quantum simulations.
 */

/**
 * @file hamiltonian.c
 *
 * @brief Functions for generating hamiltonians in matrix form. 
 * 
 * @author Robert Johansson <robert@riken.jp>
 *
 */

#include <stdio.h>
#include <math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

#include "quantum_system.h"
#include "operators.h"
#include "hamiltonian.h"
#include "param.h"
#include "misc.h"
#include "gsl_ext.h"


/**
 * \breif	Generate the hamiltonian at time t.
 *
 * For cases where the time-dependence of the Hamiltonian takes the form
 * H = H0 + H1 * f(t), this function is used to implement the specific
 * time-dependence in f(t).
 *
 * @param	H_t	Matrix where the resulting time-dependent Hamiltonian
 * 			at time t should be stored (output from this function).
 * @param	H0	The constant part of the Hamiltonian.
 * @param	H1	The part of the Hamiltonian which is modulated by f(t).
 * @param	t	The instant of time for which we should calculate H(t).
 * @param	p	Pointer to the data structure that contains all the 
 * 			simulation parameters.
 */
int
hamiltonian_t(gsl_matrix_complex *H_t, gsl_matrix_complex *H0, gsl_matrix_complex *H1, double t, solver_param *p)
{
    gsl_matrix_complex_memcpy(H_t, H0);
    return 0;

	if (p->f_A == 0.0 || t < 100.0)
	{
		// p->H_t = p->H_0;

        if (t == p->Ti)
    	    gsl_matrix_complex_memcpy(H_t, H0);
	}
	else 
	{
		// f(t) = A * sin(w * t)

		gsl_complex z;

       // printf("%s: t = %f: %f\n", __PRETTY_FUNCTION__, t, p->f_A * sin(p->f_w * t));

		gsl_matrix_complex_memcpy(H_t, H1);
		GSL_SET_COMPLEX(&z, p->f_A * sin(p->f_w * t), 0.0);
		gsl_matrix_complex_scale(H_t, z);
		gsl_matrix_complex_add(H_t, H0);
	}

    //printf("H(t = %f) =\n", t);
	//misc_matrix_complex_print_real(H_t, 0);
	
	return 0;
}

/**
 * \breif	Generates the hamiltonian for the driving field.
 *
 * Often time-dependet hamiltonians take the form H = H_0 + H_drive * f(t).
 * This function should generate the H_drive part of the total Hamiltonian.
 * @see hamiltonian_t.
 * 
 * @param	qs	Pointer to the data structure that defines
 * 			the arrangement of the quantum system.
 * @param	param	Pointer to a solver parameter structure	printf("Na =\n"); misc_matrix_complex_print_real(na, ns);
 * 			that includes all the parameter values
 * 			for the simulation.
 */
int
hamiltonian_drive(gsl_matrix_complex *H_drive, quantum_system *qs, solver_param *param)
{	
	int ns;

	ns = quantum_system_nstates(qs);

	return operator_sigma_x(H_drive, qs, 0);
}

int 
hamiltonian_z_drive(gsl_matrix_complex *H_drive, quantum_system *qs, solver_param *param)
{
        int ns;

        ns = quantum_system_nstates(qs);

        return operator_sigma_z(H_drive, qs, 0);
}

int
hamiltonian_x_drive(gsl_matrix_complex *H_drive, quantum_system *qs, solver_param *param)
{
        int ns;

        ns = quantum_system_nstates(qs);

        return operator_sigma_x(H_drive, qs, 0);
}

/*
 * Generate the hamiltonian for a N-TLS system
 *
 */
int
hamiltonian_nTLS(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
    int ns, i, j;
    gsl_matrix_complex *s1, *s2, *s3;

    ns = quantum_system_nstates(qs);

	//printf("hamiltonian_nTLS: DEBUG: Generating nTLS hamiltonian with eps = %f, %f, ...\n", param->epsilon[0]/(2*M_PI), param->epsilon[1]/(2*M_PI));

//        if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
//        {
//                fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
//                return NULL;
//        }

    gsl_matrix_complex_set_zero(H);

    s1 = gsl_matrix_complex_alloc(ns,ns);
    s2 = gsl_matrix_complex_alloc(ns,ns);
    s3 = gsl_matrix_complex_alloc(ns,ns);

    /*
     * the TLS energies
     */
    for (i = 0; i < qs->nsubsys; i++)
    {
	    gsl_complex z;

        operator_sigma_z(s1, qs, i);
		GSL_SET_COMPLEX(&z, -0.5 * param->epsilon[i], 0.0);
        gsl_matrix_complex_scale(s1, z);
        gsl_matrix_complex_add(H, s1);

        operator_sigma_x(s1, qs, i);
		GSL_SET_COMPLEX(&z, -0.5 * param->delta[i], 0.0);
        gsl_matrix_complex_scale(s1, z);
        gsl_matrix_complex_add(H, s1);
    }

        /*
         * TLS interactions
         */
        for (i = 0; i < qs->nsubsys; i++)
        {
                for (j = 0; j < qs->nsubsys; j++)
                {
                        if (param->lambda[i][j] != 0.0)
                        {
				gsl_complex z;

                                operator_sigma_x(s1, qs, i);
                                operator_sigma_x(s2, qs, j);

				GSL_SET_COMPLEX(&z, -0.5 * param->lambda[i][j], 0.0);
                                gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, z, s1, s2, GSL_COMPLEX_ZERO, s3);
                                gsl_matrix_complex_add(H, s3);
                        }
                }
        }

        gsl_matrix_complex_free(s1);
        gsl_matrix_complex_free(s2);
        gsl_matrix_complex_free(s3);

        return 0;
}


/*
 * Generate the hamiltonian for a N-TLS system 
 *
 */
int 
hamiltonian_spinchain(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns, i, j;
	gsl_matrix_complex *s1, *s2, *s3, *s1b, *s2b;

	ns = quantum_system_nstates(qs);
	
//	if ((H = gsl_matrix_complex_alloc(ns, ns)) == NULL)
//	{
//		fprintf(stderr, "hamiltonian_nTLS: Failed to allocate H. Out of memory?\n");
//		return NULL;
//	}

	gsl_matrix_complex_set_zero(H);
	
	s1  = gsl_matrix_complex_alloc(ns,ns);
	s1b = gsl_matrix_complex_alloc(ns,ns);
	s2  = gsl_matrix_complex_alloc(ns,ns);
	s2b = gsl_matrix_complex_alloc(ns,ns);
	s3  = gsl_matrix_complex_alloc(ns,ns);

	/*
	 * the TLS energies
	 */
	for (i = 0; i < qs->nsubsys; i++)
	{
		gsl_complex z;
		operator_sigma_z(s1, qs, i);
		GSL_SET_COMPLEX(&z, -0.5 * param->epsilon[i], 0.0);
		gsl_matrix_complex_scale(s1, z);
		gsl_matrix_complex_add(H, s1);
				
		operator_sigma_x(s1, qs, i);
		GSL_SET_COMPLEX(&z, -0.5 * param->delta[i], 0.0);
		gsl_matrix_complex_scale(s1, z);
		gsl_matrix_complex_add(H, s1);
	}

	/*
	 * TLS interactions
	 */
	for (i = 0; i < qs->nsubsys; i++)
	{
		for (j = 0; j < qs->nsubsys; j++)
		{
			if (param->lambda[i][j] != 0.0)
			{
				gsl_complex z;
				double theta;
				
				operator_sigma_x(s1,  qs, i);
				operator_sigma_z(s1b, qs, i);
				operator_sigma_x(s2,  qs, j);
				operator_sigma_z(s2b, qs, j);
				
				theta = tan(param->delta[i] / param->epsilon[i]);
				GSL_SET_COMPLEX(&z, sin(theta), 0.0);
				gsl_matrix_complex_scale(s1, z);
				GSL_SET_COMPLEX(&z, cos(theta), 0.0);
				gsl_matrix_complex_scale(s1b, z);
				gsl_matrix_complex_add(s1, s1b);
				theta = tan(param->delta[j] / param->epsilon[j]);
				GSL_SET_COMPLEX(&z, sin(theta), 0.0);
				gsl_matrix_complex_scale(s2, z);
				GSL_SET_COMPLEX(&z, cos(theta), 0.0);
				gsl_matrix_complex_scale(s2b, z);
				gsl_matrix_complex_add(s2, s2b);
				
				GSL_SET_COMPLEX(&z, -0.5 * param->lambda[i][j], 0.0);
				gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, z, s1, s2, GSL_COMPLEX_ZERO, s3);
				
				gsl_matrix_complex_add(H, s3);				
			}
		}
	}

	gsl_matrix_complex_free(s3);
	gsl_matrix_complex_free(s1);
	gsl_matrix_complex_free(s1b);
	gsl_matrix_complex_free(s2);
	gsl_matrix_complex_free(s2b);
	
	return 0;

}

/**
 * \breif	Generate the hamiltonian for two coupled harmonic oscillators
 *
 * This function allocate and initiates a matrix representation of a hamiltonian
 * of two coupled harmonic oscillators.
 *
 * @param	qs	Data structure that defines the quantum system.
 * @param	param	Data structure that contains all the parameters.
 *
 */
int 
hamiltonian_2ho(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *a, *ad, *b, *bd, *ws, *Hint;
	gsl_complex alpha, beta, z;
	
	ns = quantum_system_nstates(qs);
	
	gsl_matrix_complex_set_zero(H);

	Hint = gsl_matrix_complex_alloc(ns, ns);

	a  = gsl_matrix_complex_alloc(ns, ns);
	ad = gsl_matrix_complex_alloc(ns, ns);
	b  = gsl_matrix_complex_alloc(ns, ns);
	bd = gsl_matrix_complex_alloc(ns, ns);
	ws = gsl_matrix_complex_alloc(ns, ns);
	
	/* h_ w1 (ad a + 1/2) */
	operator_ho_lowering(a, qs, 0, 0);
	operator_ho_raising(ad, qs,0, 0);
	GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
	GSL_SET_COMPLEX(&beta, 0.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, beta, H);
	operator_unit(ws, qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(H, ws);
	
	/* h_ w (bd b + 1/2) */
	operator_ho_lowering(b, qs, 1, 0);
	operator_ho_raising(bd, qs, 1, 0);
	GSL_SET_COMPLEX(&alpha, param->ho_w[1], 0.0);

	GSL_SET_COMPLEX(&beta, 1.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, bd, b, beta, H);
	operator_unit(ws, qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[1], 0.0);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(H, ws);
	
	// g (a + ad) (b + bd)
	gsl_matrix_complex_add(a, ad);
	gsl_matrix_complex_add(b, bd);
	GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	GSL_SET_COMPLEX(&beta, 1.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, b, beta, H);

	// g (a + ad) (b + bd)^2
	//gsl_matrix_complex_add(a, ad);
	//gsl_matrix_complex_add(b, bd);
	// g a * b^2
	//GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	//GSL_SET_COMPLEX(&beta,  0.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, b, b, beta, Hint);
	//GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	//GSL_SET_COMPLEX(&beta,  1.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, Hint, beta, H);
	//gsl_matrix_complex_scale(Hint, );
	//gsl_matrix_complex_add(H, Hint);


	// g (a + ad)^2 (b + bd)
	//gsl_matrix_complex_add(a, ad);
	//gsl_matrix_complex_add(b, bd);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, a, a, GSL_COMPLEX_ZERO, Hint);
	//GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, Hint, b, GSL_COMPLEX_ONE, H);
	
	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	gsl_matrix_complex_free(ws);
	gsl_matrix_complex_free(Hint);

	return 0;
}


/**
 * Generate the hamiltonian for the 2-HO driving field
 *
 */
int
hamiltonian_2ho_drive(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *a, *ad; //, *b, *bd;
	
	ns = quantum_system_nstates(qs);
	
	a  = gsl_matrix_complex_alloc(ns, ns);
	ad = gsl_matrix_complex_alloc(ns, ns);

	//gsl_matrix_complex_set_zero(H);
	
	operator_ho_lowering(a, qs, 0, 0);
	operator_ho_raising(ad, qs, 0, 0);
	//b  = operator_ho_lowering(qs, 1);
	//bd = operator_ho_raising(qs, 1);

	// a bd + ad b
	//GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	//GSL_SET_COMPLEX(&beta,  0.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, bd, beta, H);
	//GSL_SET_COMPLEX(&beta,  1.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, b, beta, H);
	
	// a + ad
	gsl_matrix_complex_set_zero(H);
	gsl_matrix_complex_add(H, a);
	gsl_matrix_complex_add(H, ad);
	
	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	//gsl_matrix_complex_free(b);
	//gsl_matrix_complex_free(bd);
	
	return 0;
}

/* --------------------------------------------------------------------------------------------------
 * OPTO-MECHANIAL SYSTEM: 2HO + nonlinear interaction
 *
 */

/**
 * \breif	Generate the hamiltonian for two coupled harmonic oscillators
 *
 * This function allocate and initiates a matrix representation of a hamiltonian
 * of two coupled harmonic oscillators.
 *
 * @param	qs	Data structure that defines the quantum system.
 * @param	param	Data structure that contains all the parameters.
 *
 */
int 
hamiltonian_2ho_optmech(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *a, *ad, *na, *b, *bd, *nb, *ws, *Hint;
	gsl_complex alpha, beta, z;
	
	ns = quantum_system_nstates(qs);
	
	gsl_matrix_complex_set_zero(H);

	Hint = gsl_matrix_complex_alloc(ns, ns);

	a  = gsl_matrix_complex_alloc(ns, ns);
	ad = gsl_matrix_complex_alloc(ns, ns);
	b  = gsl_matrix_complex_alloc(ns, ns);
	bd = gsl_matrix_complex_alloc(ns, ns);
	ws = gsl_matrix_complex_alloc(ns, ns);
    na = gsl_matrix_complex_alloc(ns, ns);
    nb = gsl_matrix_complex_alloc(ns, ns);
    
	/* --- h_ w1 (ad a + 1/2) --- */
    printf("%s: ho_w[0] = %f [offset = %d]\n", __PRETTY_FUNCTION__, param->ho_w[0], param->n_offset);
	GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);

	operator_ho_lowering(a, qs, 0, 0);
	operator_ho_raising(ad, qs, 0, 0);
    operator_ho_N(na, qs, 0, param->n_offset);
	//printf("Na =\n"); misc_matrix_complex_print_real(na, ns);

	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ZERO, H);
//	gsl_matrix_complex_scale(na, alpha);
//	gsl_matrix_complex_add(H, na);

	/* --- h_ w (bd b + 1/2) --- */
    printf("%s: ho_w[1] = %f\n", __PRETTY_FUNCTION__, param->ho_w[1]);
	GSL_SET_COMPLEX(&alpha, param->ho_w[1], 0.0);

	operator_ho_lowering(b, qs, 1, 0);
	operator_ho_raising(bd, qs, 1, 0);
    operator_ho_N(nb, qs, 1, param->n_offset);
	//printf("Nb =\n"); misc_matrix_complex_print_real(nb, ns);

	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, bd, b, GSL_COMPLEX_ONE, H);
//	gsl_matrix_complex_scale(nb, alpha);
//	gsl_matrix_complex_add(H, nb);
    

	//operator_unit(ws, qs, -1);
	//GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[1], 0.0);
	//gsl_matrix_complex_scale(ws, z);
	//gsl_matrix_complex_add(H, ws);
	
	// g (a + ad) (b + bd)
	//gsl_matrix_complex_add(a, ad);
	//gsl_matrix_complex_add(b, bd);
	//GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, b, GSL_COMPLEX_ONE, H);

	// g (a + ad) (b + bd)^2
	//gsl_matrix_complex_add(a, ad);
	//gsl_matrix_complex_add(b, bd);
	// g a * b^2
	//GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	//GSL_SET_COMPLEX(&beta,  0.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, b, b, beta, Hint);
	//GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	//GSL_SET_COMPLEX(&beta,  1.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, Hint, beta, H);
	//gsl_matrix_complex_scale(Hint, );
	//gsl_matrix_complex_add(H, Hint);


	// g (a + ad)^2 (b + bd)
	//gsl_matrix_complex_add(a, ad);
	//gsl_matrix_complex_add(b, bd);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, a, a, GSL_COMPLEX_ZERO, Hint);
	//GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, Hint, b, GSL_COMPLEX_ONE, H);
	//gsl_matrix_complex_free(Hint);

    printf("%s: ho_g = %f\n", __PRETTY_FUNCTION__, param->ho_g);
	// g (ad a) (b + bd) = g na (b + bd)
	//gsl_matrix_complex_add(a, ad);
	gsl_matrix_complex_add(b, bd);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, a, a, GSL_COMPLEX_ZERO, Hint);
	GSL_SET_COMPLEX(&alpha, param->ho_g/2.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, na, b, GSL_COMPLEX_ONE, H);
	

    /* --- driving term in the rotating frame --- */
    printf("%s: ho_wd = f_A = %f\n", __PRETTY_FUNCTION__, param->f_A);
	GSL_SET_COMPLEX(&alpha, param->f_A, 0.0);
	gsl_matrix_complex_add(a, ad);
	gsl_matrix_complex_scale(a, alpha);
	gsl_matrix_complex_add(H, a);

	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	gsl_matrix_complex_free(na);
	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	gsl_matrix_complex_free(nb);
	gsl_matrix_complex_free(ws);
	gsl_matrix_complex_free(Hint);

	return 0;
}


/* --------------------------------------------------------------------------------------------------
 * OPTO-MECHANIAL SYSTEM: 2HO + nonlinear interaction? : for sideband cooling calculations
 *
 */

/**
 * \breif	Generate the hamiltonian for two coupled harmonic oscillators
 *
 * This function allocate and initiates a matrix representation of a hamiltonian
 * of two coupled harmonic oscillators.
 *
 * @param	qs	Data structure that defines the quantum system.
 * @param	param	Data structure that contains all the parameters.
 *
 */
int 
hamiltonian_2ho_cooling(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *a, *ad, *na, *b, *bd, *nb, *ws, *Hint;
	gsl_complex alpha, beta, z;
	
	ns = quantum_system_nstates(qs);
	
	gsl_matrix_complex_set_zero(H);

	Hint = gsl_matrix_complex_alloc(ns, ns);

	a  = gsl_matrix_complex_alloc(ns, ns);
	ad = gsl_matrix_complex_alloc(ns, ns);
	b  = gsl_matrix_complex_alloc(ns, ns);
	bd = gsl_matrix_complex_alloc(ns, ns);
	ws = gsl_matrix_complex_alloc(ns, ns);
    na = gsl_matrix_complex_alloc(ns, ns);
    nb = gsl_matrix_complex_alloc(ns, ns);
    
	/* --- h_ w1 (ad a + 1/2) --- */
    printf("%s: ho_w[0] = %f [offset = %d]\n", __PRETTY_FUNCTION__, param->ho_w[0], param->n_offset);
	GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);

	operator_ho_lowering(a, qs, 0, 0);
	operator_ho_raising(ad, qs, 0, 0);
    operator_ho_N(na, qs, 0, param->n_offset);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ZERO, H);

	/* --- h_ w2 (bd b + 1/2) --- */
    printf("%s: ho_w[1] = %f\n", __PRETTY_FUNCTION__, param->ho_w[1]);
	GSL_SET_COMPLEX(&alpha, param->ho_w[1], 0.0);

	operator_ho_lowering(b, qs, 1, 0);
	operator_ho_raising(bd, qs, 1, 0);
    operator_ho_N(nb, qs, 1, param->n_offset);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, bd, b, GSL_COMPLEX_ONE, H);


    /* --- interaction --- */

    // g (a + ad) (b + bd)
	//gsl_matrix_complex_add(a, ad);
	//gsl_matrix_complex_add(b, bd);
	//GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, b, GSL_COMPLEX_ONE, H);

	// g (a + ad) (b + bd)^2
	//gsl_matrix_complex_add(a, ad);
    //gsl_matrix_complex_add(b, bd);
	// g a * b^2
	//GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	//GSL_SET_COMPLEX(&beta,  0.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, b, b, GSL_COMPLEX_ZERO, Hint);
	//GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	//GSL_SET_COMPLEX(&beta,  1.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, Hint, beta, H);
	//gsl_matrix_complex_scale(Hint, );
	//gsl_matrix_complex_add(H, Hint);


	// g (a + ad)^2 (b + bd)
	gsl_matrix_complex_add(a, ad);
	gsl_matrix_complex_add(b, bd);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, a, a, GSL_COMPLEX_ZERO, Hint);
	GSL_SET_COMPLEX(&alpha, param->ho_g, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, Hint, b, GSL_COMPLEX_ONE, H);

	// g (ad a) (b + bd) = g na (b + bd)
	//gsl_matrix_complex_add(b, bd);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, ad, a, GSL_COMPLEX_ZERO, Hint);
	//GSL_SET_COMPLEX(&alpha, param->ho_g/2.0, 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, Hint, b, GSL_COMPLEX_ONE, H);
	

    /* --- driving term in the rotating frame --- */
    printf("%s: ho_wd = %f\n", __PRETTY_FUNCTION__, param->ho_wd);
	GSL_SET_COMPLEX(&alpha, param->f_A, 0.0);
	gsl_matrix_complex_add(a, ad);
	gsl_matrix_complex_scale(a, alpha);
	gsl_matrix_complex_add(H, a);

	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	gsl_matrix_complex_free(na);
	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	gsl_matrix_complex_free(nb);
	gsl_matrix_complex_free(ws);
	gsl_matrix_complex_free(Hint);

	return 0;
}




/* ---------------------------------------------------------------------------------------------------
 * Qubit + Cavity + Transmission line
 *
 */

/*
 * Generate the hamiltonian for the qubit + 2-HO driving field
 *
 */
int
hamiltonian_q2ho_drive(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *b, *bd;
	
	ns = quantum_system_nstates(qs);
	
	b  = gsl_matrix_complex_alloc(ns, ns);
	bd = gsl_matrix_complex_alloc(ns, ns);	

	operator_ho_lowering(b, qs, 2, 0);
	operator_ho_raising(bd, qs, 2, 0);

	// a + ad
	gsl_matrix_complex_set_zero(H);
	gsl_matrix_complex_add(H, b);
	gsl_matrix_complex_add(H, bd);

	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	
	return 0;
}


/*
 * Generate the hamiltonian for the Qubit + Cavity + Transmission line system/
 *
 */
int
hamiltonian_q2ho(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *a, *ad, *b, *bd, *ws, *s1;
	gsl_complex alpha, z;
	
	ns = quantum_system_nstates(qs);
	
	a  = gsl_matrix_complex_alloc(ns, ns);
	ad = gsl_matrix_complex_alloc(ns, ns);
	b  = gsl_matrix_complex_alloc(ns, ns);
	bd = gsl_matrix_complex_alloc(ns, ns);
	ws = gsl_matrix_complex_alloc(ns, ns);
	s1 = gsl_matrix_complex_alloc(ns, ns);

	gsl_matrix_complex_set_zero(H);

	/* qubit */
        operator_sigma_z(s1, qs, 0);
	GSL_SET_COMPLEX(&alpha, -0.5 * param->epsilon[0], 0.0);
        gsl_matrix_complex_scale(s1, alpha);
        gsl_matrix_complex_add(H, s1);

        operator_sigma_x(s1, qs, 0);
        GSL_SET_COMPLEX(&alpha, -0.5 * param->delta[0], 0.0);
        gsl_matrix_complex_scale(s1, alpha);
        gsl_matrix_complex_add(H, s1);
	
	/* h_ w1 (ad a + 1/2) */
	operator_ho_lowering(a, qs, 1, 0);
	operator_ho_raising(ad, qs,1, 0);
	GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ONE, H);
	operator_unit(ws, qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(H, ws);
	
	/* qubit - cavity interaction */
	operator_sigma_z(s1, qs, 0);
	gsl_matrix_complex_memcpy(ws, a);
	gsl_matrix_complex_add(ws, ad);
	GSL_SET_COMPLEX(&alpha, param->g_qc, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, s1, ws, GSL_COMPLEX_ONE, H);	

	/* h_ w (bd b + 1/2) */
	operator_ho_lowering(b, qs, 2, 0);
	operator_ho_raising(bd, qs, 2, 0);
	GSL_SET_COMPLEX(&alpha, param->ho_w[1], 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, bd, b, GSL_COMPLEX_ONE, H);
	operator_unit(ws, qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[1], 0.0);
	gsl_matrix_complex_scale(ws, z);
	gsl_matrix_complex_add(H, ws);
		
	/* cavity - transmission line interaction : g_cr (a + ad) (b + bd) * /
	gsl_matrix_complex_add(a, ad);
	gsl_matrix_complex_add(b, bd);
	GSL_SET_COMPLEX(&alpha, param->g_cr, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, a, b, GSL_COMPLEX_ONE, H);
	*/
	
	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	gsl_matrix_complex_free(ws);
	gsl_matrix_complex_free(s1);

	return 0;
}


/* -------------
 * Qubit + Cavity 
 *
 */

/*
 * Generate the hamiltonian for the qubit + HO + driving field
 *
 */
int
hamiltonian_qho_drive(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *b, *bd;
	
	ns = quantum_system_nstates(qs);
	
	b  = gsl_matrix_complex_alloc(ns, ns);
	bd = gsl_matrix_complex_alloc(ns, ns);
	
	operator_ho_lowering(b, qs, 1, param->n_offset);
	operator_ho_raising(bd, qs, 1, param->n_offset);

	// a + ad
	gsl_matrix_complex_set_zero(H);
	gsl_matrix_complex_add(H, b);
	gsl_matrix_complex_add(H, bd);

	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	
	return 0;
}

/*
 * Generate the hamiltonian for the Qubit + Cavity
 *
 */
int
hamiltonian_qho(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *a, *ad, *ws, *s1, *n;
	gsl_complex z, alpha;
	
	ns = quantum_system_nstates(qs);
	
	a  = gsl_matrix_complex_alloc(ns, ns);
	ad = gsl_matrix_complex_alloc(ns, ns);
	ws = gsl_matrix_complex_alloc(ns, ns);
	s1 = gsl_matrix_complex_alloc(ns, ns);
	n  = gsl_matrix_complex_alloc(ns, ns);

	gsl_matrix_complex_set_zero(H);

	/* qubit */
    operator_sigma_z(s1, qs, 0);
	GSL_SET_COMPLEX(&alpha, -0.5 * param->epsilon[0], 0.0);
    gsl_matrix_complex_scale(s1, alpha);
    gsl_matrix_complex_add(H, s1);

    operator_sigma_x(s1, qs, 0);
    GSL_SET_COMPLEX(&alpha, -0.5 * param->delta[0], 0.0);
    gsl_matrix_complex_scale(s1, alpha);
    gsl_matrix_complex_add(H, s1);
	
	/* h_ w1 (ad a + 1/2) */
	//a  = operator_ho_lowering(qs, 1, param->n_offset);
	//ad = operator_ho_raising(qs, 1, param->n_offset);
	//GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ONE, H);
	operator_unit(ws, qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(ws, z);
	
	operator_ho_N(n, qs, 1, param->n_offset);
	GSL_SET_COMPLEX(&z, param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(n, z);
	gsl_matrix_complex_add(ws, n);
	
	gsl_matrix_complex_add(H, ws);

	/* qubit - cavity interaction */
	operator_ho_lowering(a, qs, 1, param->n_offset);
	operator_ho_raising(ad, qs, 1, param->n_offset);
	operator_sigma_x(s1, qs, 0);
	gsl_matrix_complex_memcpy(ws, a);
	gsl_matrix_complex_add(ws, ad);
	GSL_SET_COMPLEX(&alpha, param->g_qc, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, s1, ws, GSL_COMPLEX_ONE, H);

	gsl_matrix_complex_free(ws);
	gsl_matrix_complex_free(s1);	
	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	gsl_matrix_complex_free(n);

	return 0;
}

/* ---------------------
 * 3-level system + Cavity 
 *
 */

/*
 * Generate the hamiltonian for the 3-level atom + HO + driving field
 *
 */
int
hamiltonian_3ls_ho_drive(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *b, *bd;
	
	ns = quantum_system_nstates(qs);
	
	b  = gsl_matrix_complex_alloc(ns, ns);
	bd = gsl_matrix_complex_alloc(ns, ns);
	
	operator_ho_lowering(b, qs, 1, param->n_offset);
	operator_ho_raising(bd, qs, 1, param->n_offset);

	// a + ad
	gsl_matrix_complex_set_zero(H);
	gsl_matrix_complex_add(H, b);
	gsl_matrix_complex_add(H, bd);

	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	
	return 0;
}

/*
 * Generate the hamiltonian for the Qubit + Cavity
 *
 */
int
hamiltonian_3ls_ho(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *a, *ad, *ws, *s1, *n;
	gsl_complex z, alpha;
	
	ns = quantum_system_nstates(qs);
	
	a  = gsl_matrix_complex_alloc(ns, ns);
	ad = gsl_matrix_complex_alloc(ns, ns);
	ws = gsl_matrix_complex_alloc(ns, ns);
	s1 = gsl_matrix_complex_alloc(ns, ns);
	n  = gsl_matrix_complex_alloc(ns, ns);

	gsl_matrix_complex_set_zero(H);

	/* qubit */
        operator_3ls_sigma_z(s1, qs, 0);
	GSL_SET_COMPLEX(&alpha, -0.5 * param->epsilon[0], 0.0);
        gsl_matrix_complex_scale(s1, alpha);
        gsl_matrix_complex_add(H, s1);
        
        operator_3ls_sigma_x(s1, qs, 0);
        GSL_SET_COMPLEX(&alpha, -0.5 * param->delta[0], 0.0);
        gsl_matrix_complex_scale(s1, alpha);
        gsl_matrix_complex_add(H, s1);

	/* h_ w1 (ad a + 1/2) */
	//a  = operator_ho_lowering(qs, 1, param->n_offset);
	//ad = operator_ho_raising(qs, 1, param->n_offset);
	//GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
	//gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ONE, H);
	operator_unit(ws, qs, -1);
	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(ws, z);
	
	operator_ho_N(n, qs, 1, param->n_offset);
	GSL_SET_COMPLEX(&z, param->ho_w[0], 0.0);
	gsl_matrix_complex_scale(n, z);
	gsl_matrix_complex_add(ws, n);
	
	gsl_matrix_complex_add(H, ws);

	/* qubit - cavity interaction */
	operator_ho_lowering(a, qs, 1, param->n_offset);
	operator_ho_raising(ad, qs, 1, param->n_offset);
	operator_3ls_sigma_z(s1, qs, 0);
	gsl_matrix_complex_memcpy(ws, a);
	gsl_matrix_complex_add(ws, ad);
	GSL_SET_COMPLEX(&alpha, param->g_qc, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, s1, ws, GSL_COMPLEX_ONE, H);

	gsl_matrix_complex_free(ws);
	gsl_matrix_complex_free(s1);
	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	gsl_matrix_complex_free(n);

	return 0;
}

/*
 * Generate the hamiltonian for the HO + driving field
 *
 */
int
hamiltonian_ho_drive(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *b, *bd;
	
	ns = quantum_system_nstates(qs);
	
	b  = gsl_matrix_complex_alloc(ns, ns);
	bd = gsl_matrix_complex_alloc(ns, ns);
	
	operator_ho_lowering(b, qs, 0, 0);
	operator_ho_raising(bd, qs, 0, 0);

	// a + ad
	gsl_matrix_complex_set_zero(H);
	gsl_matrix_complex_add(H, b);
	gsl_matrix_complex_add(H, bd);
	
	gsl_matrix_complex_free(b);
	gsl_matrix_complex_free(bd);
	
	return 0;
}

/*
 * Generate the hamiltonian for the harmonic oscillator
 *
 */
int
hamiltonian_ho(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *ws, *n; //*a, *ad, *ws;
	gsl_complex z;
	
	ns = quantum_system_nstates(qs);
	
	n  = gsl_matrix_complex_alloc(ns, ns);
	ws = gsl_matrix_complex_alloc(ns, ns);

	gsl_matrix_complex_set_zero(H);

        /* h_ w1 (ad a + 1/2) */
//        a  = operator_ho_lowering(qs, 1, param->n_offset);
//        ad = operator_ho_raising(qs, 1, param->n_offset);
//        GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
//        gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ONE, H);
        operator_unit(ws, qs, -1);
        GSL_SET_COMPLEX(&z, 0.0 * param->ho_w[0], 0.0);
        gsl_matrix_complex_scale(ws, z);

        operator_ho_N(n, qs, 0, param->n_offset);
        GSL_SET_COMPLEX(&z, param->ho_w[0], 0.0);
        gsl_matrix_complex_scale(n, z);
        gsl_matrix_complex_add(ws, n);

        gsl_matrix_complex_add(H, ws);
        
	gsl_matrix_complex_free(ws);
        gsl_matrix_complex_free(n);

//	/* h_ w1 (ad a + 1/2) */
//	a  = operator_ho_lowering(qs, 0, 0);
//	ad = operator_ho_raising(qs, 0, 0);
//	GSL_SET_COMPLEX(&alpha, param->ho_w[0], 0.0);
//	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, GSL_COMPLEX_ONE, H);
//	ws = operator_unit(qs, -1);
//	GSL_SET_COMPLEX(&z, 0.5 * param->ho_w[0], 0.0);
//	gsl_matrix_complex_scale(ws, z);
//	gsl_matrix_complex_add(H, ws);
//	gsl_matrix_complex_free(ws);
//
//	gsl_matrix_complex_free(a);
//	gsl_matrix_complex_free(ad);

	return 0;
}

/* ---------------------------------------------------------------------------------------
 * 3LS + large spin + harmonic oscillator
 *
 */

/*
 * Generate the dicke model hamiltonian + 3LS 
 *
 */
int
hamiltonian_dicke(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *Hsub, *n, *sx, *Jp, *Jm, *J_sum, *a, *ad, *a_sum;
	gsl_complex z;
	
	ns = quantum_system_nstates(qs);
	
	Hsub  = gsl_matrix_complex_alloc(ns, ns);
	n     = gsl_matrix_complex_alloc(ns, ns);
	sx    = gsl_matrix_complex_alloc(ns, ns);
	Jp    = gsl_matrix_complex_alloc(ns, ns);
	Jm    = gsl_matrix_complex_alloc(ns, ns);
	J_sum = gsl_matrix_complex_alloc(ns, ns);
	a     = gsl_matrix_complex_alloc(ns, ns);
	ad    = gsl_matrix_complex_alloc(ns, ns);
	a_sum = gsl_matrix_complex_alloc(ns, ns);

	/* 
 	 * start by blanking the hamiltonian matrix
 	 */
	gsl_matrix_complex_set_zero(H);

	/*
 	 * subsystem 1: 3LS artificial atom: 
 	 *
 	 * /
        Hsub = operator_sigma_z(qs, 0);
        GSL_SET_COMPLEX(&z, 0.5 * param->epsilon[0], 0.0);
        gsl_matrix_complex_scale(Hsub, z);
        gsl_matrix_complex_add@mainpage(H, Hsub);
        gsl_matrix_complex_free(Hsub);

        Hsub = operator_sigma_x(qs, 0);
        GSL_SET_COMPLEX(&z, 0.5 * param->delta[0], 0.0);
        gsl_matrix_complex_scale(Hsub, z);
        gsl_matrix_complex_add(H, Hsub);
        gsl_matrix_complex_free(Hsub);
	*/

	/* 
 	 * subsystem 2: N/2 spin:
 	 *
 	 */
	operator_Jz(Hsub, qs, 1-1, param->j);
        GSL_SET_COMPLEX(&z, param->J_w[0], 0.0);
        gsl_matrix_complex_scale(Hsub, z);
	gsl_matrix_complex_add(H, Hsub);

	/* 
	 * subsystem 3: HO: h_ w1 (ad a + 1/2) 
	 *
	 */
        operator_ho_N(Hsub, qs, 2-1, param->n_offset);
        GSL_SET_COMPLEX(&z, -param->ho_w[0], 0.0);
        gsl_matrix_complex_scale(Hsub, z);
        gsl_matrix_complex_add(H, Hsub);

	/*
 	 * Interaction: 
 	 * 1) artificial atom <-> oscillator
 	 * 2) spin <-> oscillator
 	 *
 	 */
        operator_ho_lowering(a, qs, 2-1, param->n_offset);
        operator_ho_raising(ad, qs, 2-1, param->n_offset);
        gsl_matrix_complex_memcpy(a_sum, a);
        gsl_matrix_complex_add   (a_sum, ad);
        
	// 1)
	/*
	sx = operator_sigma_x(qs, 0);

        GSL_SET_COMPLEX(&z, param->g_ac, 0.0);
        gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, z, sx, a_sum, GSL_COMPLEX_ONE, H);

        gsl_matrix_complex_free(sx);
	*/
	// 2)
	
	operator_J_plus (Jp, qs, 1-1, param->j);
	operator_J_minus(Jm, qs, 1-1, param->j);
        gsl_matrix_complex_memcpy(J_sum, Jp);
        gsl_matrix_complex_add   (J_sum, Jm);

        GSL_SET_COMPLEX(&z, param->g_sc, 0.0);
        gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, z, J_sum, a_sum, GSL_COMPLEX_ONE, H);

   
        gsl_matrix_complex_free(Hsub);
        gsl_matrix_complex_free(Jp);
        gsl_matrix_complex_free(Jm);
        gsl_matrix_complex_free(J_sum);
        gsl_matrix_complex_free(a);
        gsl_matrix_complex_free(ad);
        gsl_matrix_complex_free(a_sum);
        gsl_matrix_complex_free(sx);
        gsl_matrix_complex_free(n);

	return 0;
}

//!-----------------------------------------------------------------------------
// @breif   Generate squeezing hamiltonian for a cavity mode 
//
// This function allocate and initiates a matrix representation of a hamiltonian
// for a parametric amplification: H = i eps [ (ad^\dag)^2 - a^2 ]
//
// @param	qs	Data structure that defines the quantum system.
// @param	param	Data structure that contains all the parameters.
//------------------------------------------------------------------------------
int 
hamiltonian_squeezing(gsl_matrix_complex *H, quantum_system *qs, solver_param *param)
{
	int ns;
	gsl_matrix_complex *a, *ad, *aa, *adad;
	gsl_complex z;
	
	ns = quantum_system_nstates(qs);
	
	a    = gsl_matrix_complex_alloc(ns, ns);
	ad   = gsl_matrix_complex_alloc(ns, ns);
	aa   = gsl_matrix_complex_alloc(ns, ns);
	adad = gsl_matrix_complex_alloc(ns, ns);
	
	operator_ho_lowering(a, qs, 0, 0);
	operator_ho_raising(ad, qs, 0, 0);

  	GSL_SET_COMPLEX(&z, 0.0, +param->xi);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, z, ad, ad, GSL_COMPLEX_ZERO, adad);
  	GSL_SET_COMPLEX(&z, 0.0, -param->xi);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, z,  a,  a, GSL_COMPLEX_ZERO, aa);

	gsl_matrix_complex_set_zero(H);
	gsl_matrix_complex_add(H, adad);
	gsl_matrix_complex_add(H,   aa);
	
	gsl_matrix_complex_free(a);
	gsl_matrix_complex_free(ad);
	gsl_matrix_complex_free(aa);
	gsl_matrix_complex_free(adad);

	return 0;
}


