/*
 * Solvers for the Schroding equation and master equations.
 * 
 * 
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: master_equation.c,v 1.1.1.1 2008-09-03 09:36:41 rob Exp $
 */

#include <stdio.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>

#include "quantum_system.h"
#include "operators.h"
#include "hamiltonian.h"
#include "master_equation.h"
#include "param.h"
#include "misc.h"
#include "integrate.h"
#include "gsl_ext.h"
#include "gsl_ext_expm_complex.h"
#include "basis_transform.h"

/*
 * Evolve the density matrix rho0 assuming a non-dissipative dynamics
 * and a constant hamiltonian, using a very simple method. 
 * param->Ti = start time
 * param->Tf = end time
 *
 *
 */
int
dm_evolve_unitary_const_simple(quantum_system *qs, gsl_matrix_complex *rho0,
		               solver_param *param, rho_cb_func_t rho_cb_func)
{
	gsl_matrix_complex *rho_t, *H;
	double t;
	int ns;

	ns = quantum_system_nstates(qs);
	
	rho_t = gsl_matrix_complex_alloc(ns, ns);
	H     = gsl_matrix_complex_alloc(ns, ns);

	param->H0_func(H, qs, param);

	//printf("H_real =\n");
        //misc_matrix_complex_print_real(H, ns);
	//printf("H_imag =\n");
        //misc_matrix_complex_print_imag(H, ns);
	
	gsl_matrix_complex_memcpy(rho_t, rho0);
	rho_cb_func(rho0, param->Ti, qs, param);

	for (t = param->Ti; t < param->Tf; t += param->dT)
	{
                //gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho_t, eHr, beta, H);
                //gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, eHl, H, beta, rho_t);
		
		rho_cb_func(rho_t, t, qs, param);
	}

	gsl_matrix_complex_free(rho_t);
	gsl_matrix_complex_free(H);
	
	return 0;
}

/*
 * Evolve the density matrix rho0 assuming a non-dissipative dynamics
 * and a constant hamiltonian. 
 * param->Ti = start time
 * param->Tf = end time
 *
 *
 */
int
dm_evolve_unitary_const(quantum_system *qs, gsl_matrix_complex *rho0,
		        solver_param *param, rho_cb_func_t rho_cb_func)
{
	gsl_matrix_complex *eHr, *eHl, *rho_t, *H;
	double t;
	gsl_complex alpha, beta, z;
	int ns;

	ns = quantum_system_nstates(qs);
	
	eHl = gsl_matrix_complex_alloc(ns, ns);
	eHr = gsl_matrix_complex_alloc(ns, ns);
	rho_t = gsl_matrix_complex_alloc(ns, ns);

	H     = gsl_matrix_complex_alloc(ns, ns);
	param->H0_func(H, qs, param);

	//printf("H_real =\n");
        //misc_matrix_complex_print_real(H, ns);
	//printf("H_imag =\n");
        //misc_matrix_complex_print_imag(H, ns);
	
	GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	GSL_SET_COMPLEX(&beta,  0.0, 0.0);
		
	GSL_SET_COMPLEX(&z, 0.0, -1.0 * param->dT);
	gsl_matrix_complex_scale(H, z);
#ifndef RSCC
	gsl_ext_expm_complex(H, eHl);
#else
	fprintf(stderr, "WARNING: gsl_ext_expm_complex not supported on RSCC.\n");
#endif
	//gsl_linalg_complex_exponential_ss(H, eHl, 0);
	
	GSL_SET_COMPLEX(&z, -1.0, 0.0);
	gsl_matrix_complex_scale(H, z);
#ifndef RSCC
	gsl_ext_expm_complex(H, eHr);
#else
        fprintf(stderr, "WARNING: gsl_ext_expm_complex not supported on RSCC.\n");
#endif

	//gsl_linalg_complex_exponential_ss(H, eHr, 0);

	gsl_matrix_complex_memcpy(rho_t, rho0);
	rho_cb_func(rho0, param->Ti, qs, param);

	for (t = param->Ti; t < param->Tf; t += param->dT)
	{
                gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho_t, eHr, beta, H);
                gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, eHl, H, beta, rho_t);
		
		rho_cb_func(rho_t, t, qs, param);
	}

	gsl_matrix_complex_free(eHl);
	gsl_matrix_complex_free(eHr);
	gsl_matrix_complex_free(rho_t);
	gsl_matrix_complex_free(H);
	
	return 0;
}

/*
 * Evolve the density matrix rho0 assuming a non-dissipative dynamics
 * and a constant hamiltonian. 
 * param->Ti = start time
 * param->Tf = end time
 *
 * XXX: This solver needs to be updated to use complex hamiltonian matrices
 */
int
dm_evolve_unitary_const_full_steps(quantum_system *qs, gsl_matrix_complex *rho0,
		        solver_param *param, rho_cb_func_t rho_cb_func)
{
	gsl_matrix_complex *eHr, *eHl, *rho_t, *H, *R;
	double t;
	gsl_complex alpha, beta, z;
	int ns;

	ns = quantum_system_nstates(qs);

	eHl = gsl_matrix_complex_alloc(ns, ns);
	eHr = gsl_matrix_complex_alloc(ns, ns);
	R   = gsl_matrix_complex_alloc(ns, ns);
	rho_t = gsl_matrix_complex_alloc(ns, ns);

	//H = hamiltonian_nTLS(qs, param);
	H     = gsl_matrix_complex_alloc(ns, ns);
        param->H0_func(H, qs, param);
	
	//printf("H =\n");
        //misc_matrix_complex_print(H, ns);
	//gsl_ext_matrix_convert(R, H);
        //printf("R =\n");
        //misc_matrix_complex_print(R, ns);
	
	GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	GSL_SET_COMPLEX(&beta,  0.0, 0.0);
		
	rho_cb_func(rho0, param->Ti, qs, param);

	for (t = param->Ti; t < param->Tf; t += param->dT)
	{
		gsl_matrix_complex_memcpy(R, H);
		
		GSL_SET_COMPLEX(&z, 0.0, -1.0 * (t - param->Ti));
		gsl_matrix_complex_scale(R, z);
#ifndef RSCC
		gsl_ext_expm_complex(R, eHl);
#else
        fprintf(stderr, "WARNING: gsl_ext_expm_complex not supported on RSCC.\n");
#endif
	
		GSL_SET_COMPLEX(&z, -1.0, 0.0);
		gsl_matrix_complex_scale(R, z);
#ifndef RSCC
		gsl_ext_expm_complex(R, eHr);
#endif

                gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho0, eHr, beta, R);
                gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, eHl, R, beta, rho_t);
		
		rho_cb_func(rho_t, t, qs, param);
	}

	//gsl_matrix_complex_free(H);
	//gsl_matrix_complex_free(R);
	//gsl_matrix_complex_free(eHl);
	//gsl_matrix_complex_free(eHr);
	return 0;
}



/* ---------------------------------------------------------------------------------------- 
 * Evolve the density matrix rho0 assuming a non-dissipative dynamics and time-dependent
 * hamiltonian. 
 * param->Ti = start time
 * param->Tf = end time
 *
 *
 */
int
unitary_func_drho_dt(gsl_matrix_complex *rho_t, gsl_matrix_complex *drho_dt, double t,
		     quantum_system *qs, solver_param *p)
{
	gsl_complex z, alpha, beta;
	//gsl_matrix_complex *H;

	//printf("DEBUG: unitary_func_drho_dt: t = %f\n", t);

	//hamiltonian_t(p->H_t, p->H0, p->H1, t, p);
	p->Ht_func(p->H_t, p->H0, p->H1, t, p);

	//printf("H(t = %f) =\n", t);
    //misc_matrix_complex_print(p->H_t, quantum_system_nstates(qs));
	
	// drho_dt = -i ( H * rho - rho * H )
    GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
    GSL_SET_COMPLEX(&beta,  0.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho_t, p->H_t, beta, drho_dt);
    GSL_SET_COMPLEX(&beta, -1.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->H_t, rho_t, beta, drho_dt);
	GSL_SET_COMPLEX(&z, 0.0, -1.0);
	gsl_matrix_complex_scale(drho_dt, z);
	
	return 0;
}

int
dm_evolve_unitary_t(quantum_system *qs, gsl_matrix_complex *rho0,
		    solver_param *param, rho_cb_func_t rho_cb_func)
{
	int ns;

	//printf("DEBUG: starting dm_evolve_unitary_t\n");

    ns = quantum_system_nstates(qs);

    //param->H0  = gsl_matrix_complex_alloc(ns, ns);
    //param->H1  = gsl_matrix_complex_alloc(ns, ns);
    //param->H_t = gsl_matrix_complex_alloc(ns, ns);

	param->H0_func(param->H0, qs, param);
	param->H1_func(param->H1, qs, param);
	
    //H = hamiltonian_nTLS(qs, param);
    //gsl_ext_matrix_convert(param->H0, H);
	//gsl_matrix_free(H);

    //H = hamiltonian_drive(qs, param);
    //gsl_ext_matrix_convert(param->H1, H);
	//gsl_matrix_free(H);
	
	//odeint_complex_rk(qs, param, rho0, unitary_func_drho_dt, rho_cb_func);
	odeint_complex_rk_adaptive(qs, param, rho0, unitary_func_drho_dt, rho_cb_func);

	//gsl_matrix_complex_free(param->H0);
	//gsl_matrix_complex_free(param->H1);
	//gsl_matrix_complex_free(param->H_t);

	return 0;
}

/* ---------------------------------------------------------------------------------------- 
 * Evolve the density matrix rho0 using a dissipative dynamics and time-dependent
 * hamiltonian. 
 * param->Ti = start time
 * param->Tf = end time
 *
 *
 */
int
lme_func_drho_dt(gsl_matrix_complex *rho_t, gsl_matrix_complex *drho_dt, double t,
                 quantum_system *qs, solver_param *p)
{
	int i;
	gsl_complex z, alpha, beta;

	//hamiltonian_t(p->H_t, p->H0, p->H1, t, p);
    if (p->Ht_func)
    {
        p->Ht_func(p->H_t, p->H0, p->H1, t, p);
    }
    else
    {
        p->H_t = p->H0;
    }

	// drho_dt = -i ( H * rho - rho * H )
    GSL_SET_COMPLEX(&beta,  0.0, 0.0);
    GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho_t, p->H_t, beta, drho_dt);
    GSL_SET_COMPLEX(&beta,  -1.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->H_t, rho_t, beta, drho_dt);
	GSL_SET_COMPLEX(&z, 0.0, -1.0);
	gsl_matrix_complex_scale(drho_dt, z);

	// optional continuous basis transformation
	// drho_dt = ( dS/dt S^* * rho - rho * dS/dt S^* )
	if (p->cont_basis_transform == 1)
	{
		gsl_matrix_complex *S, *dS;
		
		printf("DEBUG: using continous basis transformation of master eqaution: %f\n", t);
		S  = basis_transform(qs, p, 0, p->ho_w_func, t);
		dS = basis_transform_derivative(qs, p, 0, p->ho_w_func, t);
		
		gsl_blas_zgemm(CblasNoTrans, CblasConjTrans, GSL_COMPLEX_ONE, dS, S, GSL_COMPLEX_ZERO, p->dSSt);
		
		gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE,    rho_t, p->dSSt, GSL_COMPLEX_ONE, drho_dt);
		gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_NEGONE, p->dSSt, rho_t, GSL_COMPLEX_ONE, drho_dt);

		gsl_matrix_complex_free(S);
		gsl_matrix_complex_free(dS);
	}
	
	// loop through all dissipation operators
	// drho_t += g/2 (N+1)( 2 a rho ad - ad a rho - rho ad a) + g/2 N ( 2 ad rho a - a ad rho - rho a ad)
	for (i = 0; i < p->do_n; i++)
	{
		if (p->do_g1[i] != 0.0)
		{
			//printf("DEBUG: LME: operator %d, a. (%f, %f)\n", i, p->do_g1[i], p->do_g2[i]);
			gsl_matrix_complex_set_zero(p->ws1);
			gsl_matrix_complex_set_zero(p->ws2);
			
			GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
			
			GSL_SET_COMPLEX(&beta,  0.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho_t, p->do_ada[i], beta, p->ws1);
			
			GSL_SET_COMPLEX(&beta,  1.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->do_ada[i], rho_t, beta, p->ws1);
			
			GSL_SET_COMPLEX(&beta,  0.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->do_a[i], rho_t, beta, p->ws2);
			GSL_SET_COMPLEX(&alpha, -2.0, 0.0);
			GSL_SET_COMPLEX(&beta,  1.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasConjTrans, alpha, p->ws2, p->do_a[i], beta, p->ws1);
			
			GSL_SET_COMPLEX(&z, -p->do_g1[i] / 2, 0.0);
			gsl_matrix_complex_scale(p->ws1, z);
			
			gsl_matrix_complex_add(drho_dt, p->ws1);
		}

		if (p->do_g2[i] != 0.0)
		{
			//printf("DEBUG: LME: operator %d, ad. (%f, %f)\n", i, p->do_g1[i], p->do_g2[i]);
			gsl_matrix_complex_set_zero(p->ws1);
			gsl_matrix_complex_set_zero(p->ws2);
			
			GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
			
			GSL_SET_COMPLEX(&beta,  0.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho_t, p->do_aad[i], beta, p->ws1);
			
			GSL_SET_COMPLEX(&beta,  1.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->do_aad[i], rho_t, beta, p->ws1);
			
			GSL_SET_COMPLEX(&beta,  0.0, 0.0);
			gsl_blas_zgemm(CblasConjTrans, CblasNoTrans, alpha, p->do_a[i], rho_t, beta, p->ws2);
			GSL_SET_COMPLEX(&alpha, -2.0, 0.0);
			GSL_SET_COMPLEX(&beta,  1.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->ws2, p->do_a[i], beta, p->ws1);
			
			GSL_SET_COMPLEX(&z, -p->do_g2[i]/2, 0.0);
			gsl_matrix_complex_scale(p->ws1, z);
		
			gsl_matrix_complex_add(drho_dt, p->ws1);
		}		
	}

	return 0;
}

int
dm_evolve_lme_t(quantum_system *qs, gsl_matrix_complex *rho0,
		        solver_param *param, rho_cb_func_t rho_cb_func)
{
	gsl_complex alpha, beta;
	int ns, i;

	printf("DEBUG: starting dm_evolve_lme_t (Lindblad-type Master Eqaution)\n");

    ns = quantum_system_nstates(qs);

	/* --- calculate hamiltonians --- */
	
	// work space matrices
    param->ws1 = gsl_matrix_complex_alloc(ns, ns);
    param->ws2 = gsl_matrix_complex_alloc(ns, ns);

	param->H0_func(param->H0, qs, param);

	if (param->H1_func)
		param->H1_func(param->H1, qs, param);
	else
		param->H1 = NULL;

	if (param->cont_basis_transform == 1)
	{
		param->dSSt = gsl_matrix_complex_alloc(ns, ns);
	}

	/* --- Precalculate lindblad operators --- */
	for (i = 0; i < param->do_n; i++)
	{	
		GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	    GSL_SET_COMPLEX(&beta,  0.0, 0.0);

		param->do_aad[i] = NULL;
		param->do_ada[i] = NULL;

        if (param->do_g2[i] != 0.0)
		{
			param->do_aad[i] = gsl_matrix_complex_alloc(ns, ns);
	        gsl_blas_zgemm(CblasNoTrans, CblasConjTrans, alpha, param->do_a[i], param->do_a[i], beta, param->do_aad[i]);
						
		}
		if (param->do_g1[i] != 0.0)
		{
            param->do_ada[i] = gsl_matrix_complex_alloc(ns, ns);
	        gsl_blas_zgemm(CblasConjTrans, CblasNoTrans, alpha, param->do_a[i], param->do_a[i], beta, param->do_ada[i]);
		}
	}
	

	/* --- Evolve the system with an ODE solver --- */
	
	odeint_complex_rk_adaptive(qs, param, rho0, lme_func_drho_dt, rho_cb_func);
	
	
	/* --- free matrices --- */
	for (i = 0; i < param->do_n; i++)
	{
		if (param->do_aad[i] != NULL) gsl_matrix_complex_free(param->do_aad[i]);
		if (param->do_ada[i] != NULL) gsl_matrix_complex_free(param->do_ada[i]);
	}
	
	gsl_matrix_complex_free(param->ws1);
	gsl_matrix_complex_free(param->ws2);

	return 0;
}

/* ---------------------------------------------------------------------------------------- 
 * Evolve the density matrix rho0 using a dissipative dynamics and time-dependent
 * hamiltonian, using the instantaneous basis for dissipation operators...
 * 
 * param->Ti = start time
 * param->Tf = end time
 *
 *
 */
int
lme_ib_func_drho_dt(gsl_matrix_complex *rho_t, gsl_matrix_complex *drho_dt, double t,
                    quantum_system *qs, solver_param *p)
{
	int i;
	gsl_complex z, alpha, beta;

	//hamiltonian_t(p->H_t, p->H0, p->H1, t, p);
	p->Ht_func(p->H_t, p->H0, p->H1, t, p);

	// drho_dt = -i ( H * rho - rho * H )
    GSL_SET_COMPLEX(&beta,  0.0, 0.0);
    GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho_t, p->H_t, beta, drho_dt);
    GSL_SET_COMPLEX(&beta,  -1.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->H_t, rho_t, beta, drho_dt);
	GSL_SET_COMPLEX(&z, 0.0, -1.0);
	gsl_matrix_complex_scale(drho_dt, z);

	//
	// recalculate the dissipation operators... transform to the instantaneous
	// eigenbasis
	gsl_eigen_hermv(p->H_t, p->eval, p->evec, p->w);
	gsl_eigen_hermv_sort(p->eval, p->evec, GSL_EIGEN_SORT_VAL_ASC);

	for (i = 0; i < p->do_n; i++)
	{	
		GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	    GSL_SET_COMPLEX(&beta,  0.0, 0.0);

		// do_a_eb[i] = S do_a[i] S^{-1}
       		gsl_blas_zgemm(CblasNoTrans,   CblasNoTrans, GSL_COMPLEX_ONE, p->evec,   p->do_a[i], GSL_COMPLEX_ZERO, p->op_tmp);
       		gsl_blas_zgemm(CblasNoTrans, CblasConjTrans, GSL_COMPLEX_ONE, p->op_tmp, p->evec,    GSL_COMPLEX_ZERO, p->do_a_eb[i]);

       		gsl_blas_zgemm(CblasNoTrans,   CblasNoTrans, GSL_COMPLEX_ONE, p->evec,   p->do_ad[i], GSL_COMPLEX_ZERO, p->op_tmp);
       		gsl_blas_zgemm(CblasNoTrans, CblasConjTrans, GSL_COMPLEX_ONE, p->op_tmp, p->evec,     GSL_COMPLEX_ZERO, p->do_ad_eb[i]);


        	if (p->do_g2[i] != 0.0)
		{
	       		gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, p->evec,   p->do_aad[i], GSL_COMPLEX_ZERO, p->op_tmp);
	       		gsl_blas_zgemm(CblasNoTrans,   CblasConjTrans, GSL_COMPLEX_ONE, p->op_tmp, p->evec,     GSL_COMPLEX_ZERO, p->do_aad_eb[i]);
						
		}
		if (p->do_g1[i] != 0.0)
		{
	       		gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE, p->evec,   p->do_ada[i], GSL_COMPLEX_ZERO, p->op_tmp);
	       		gsl_blas_zgemm(CblasNoTrans,   CblasConjTrans, GSL_COMPLEX_ONE, p->op_tmp, p->evec,     GSL_COMPLEX_ZERO, p->do_ada_eb[i]);
		}
	}

	// optional continuous basis transformation
	// drho_dt = ( dS/dt S^* * rho - rho * dS/dt S^* )
	if (p->cont_basis_transform == 1)
	{
		gsl_matrix_complex *S, *dS;
		
		S  = basis_transform(qs, p, 0, p->ho_w_func, t);
		dS = basis_transform_derivative(qs, p, 0, p->ho_w_func, t);
		
		gsl_blas_zgemm(CblasNoTrans, CblasConjTrans, GSL_COMPLEX_ONE, dS, S, GSL_COMPLEX_ZERO, p->dSSt);
		
		gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_ONE,    rho_t, p->dSSt, GSL_COMPLEX_ONE, drho_dt);
		gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, GSL_COMPLEX_NEGONE, p->dSSt, rho_t, GSL_COMPLEX_ONE, drho_dt);

		gsl_matrix_complex_free(S);
		gsl_matrix_complex_free(dS);
	}
	
	// loop through all dissipation operators
	// drho_t += g/2 (N+1)( 2 a rho ad - ad a rho - rho ad a) + g/2 N ( 2 ad rho a - a ad rho - rho a ad)
	for (i = 0; i < p->do_n; i++)
	{
		if (p->do_g1[i] != 0.0)
		{
			gsl_matrix_complex_set_zero(p->ws1);
			gsl_matrix_complex_set_zero(p->ws2);
			
			GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
			
			GSL_SET_COMPLEX(&beta,  0.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho_t, p->do_ada_eb[i], beta, p->ws1);
			
			GSL_SET_COMPLEX(&beta,  1.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->do_ada_eb[i], rho_t, beta, p->ws1);
			
			GSL_SET_COMPLEX(&beta,  0.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->do_a_eb[i], rho_t, beta, p->ws2);
			GSL_SET_COMPLEX(&alpha, -2.0, 0.0);
			GSL_SET_COMPLEX(&beta,  1.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasConjTrans, alpha, p->ws2, p->do_a_eb[i], beta, p->ws1);
			
			GSL_SET_COMPLEX(&z, -p->do_g1[i] / 2, 0.0);
			gsl_matrix_complex_scale(p->ws1, z);
			
			gsl_matrix_complex_add(drho_dt, p->ws1);
		}

		if (p->do_g2[i] != 0.0)
		{
			//printf("DEBUG: LME: operator %d, ad. (%f, %f)\n", i, p->do_g1[i], p->do_g2[i]);
			gsl_matrix_complex_set_zero(p->ws1);
			gsl_matrix_complex_set_zero(p->ws2);
			
			GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
			
			GSL_SET_COMPLEX(&beta,  0.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho_t, p->do_aad_eb[i], beta, p->ws1);
			
			GSL_SET_COMPLEX(&beta,  1.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->do_aad_eb[i], rho_t, beta, p->ws1);
			
			GSL_SET_COMPLEX(&beta,  0.0, 0.0);
			gsl_blas_zgemm(CblasConjTrans, CblasNoTrans, alpha, p->do_a_eb[i], rho_t, beta, p->ws2);
			GSL_SET_COMPLEX(&alpha, -2.0, 0.0);
			GSL_SET_COMPLEX(&beta,  1.0, 0.0);
			gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->ws2, p->do_a_eb[i], beta, p->ws1);
			
			GSL_SET_COMPLEX(&z, -p->do_g2[i]/2, 0.0);
			gsl_matrix_complex_scale(p->ws1, z);
		
			gsl_matrix_complex_add(drho_dt, p->ws1);
		}		
	}

	return 0;
}

int
dm_evolve_lme_ib_t(quantum_system *qs, gsl_matrix_complex *rho0,
		   solver_param *param, rho_cb_func_t rho_cb_func)
{
	gsl_complex alpha, beta;
	int ns, i;

    ns = quantum_system_nstates(qs);

	/* --- calculate hamiltonians --- */
    //param->H_t = gsl_matrix_complex_alloc(ns, ns);
	
	// work space matrices
    param->ws1 = gsl_matrix_complex_alloc(ns, ns);
    param->ws2 = gsl_matrix_complex_alloc(ns, ns);
    param->op_tmp = gsl_matrix_complex_alloc(ns, ns);

	param->H0_func(param->H0, qs, param);
	if (param->H1_func)
		param->H1_func(param->H1, qs, param);
	else
		param->H1 = NULL;

	if (param->cont_basis_transform == 1)
	{
		param->dSSt = gsl_matrix_complex_alloc(ns, ns);
	}

	/* --- Precalculate lindblad operators --- */
	for (i = 0; i < param->do_n; i++)
	{	
		GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
	        GSL_SET_COMPLEX(&beta,  0.0, 0.0);

		param->do_ad_eb[i] = gsl_matrix_complex_alloc(ns, ns);
		param->do_a_eb[i]  = gsl_matrix_complex_alloc(ns, ns);

		param->do_aad[i] = NULL;
		param->do_ada[i] = NULL;

        	if (param->do_g2[i] != 0.0)
		{
			param->do_aad[i] = gsl_matrix_complex_alloc(ns, ns);
	        	gsl_blas_zgemm(CblasNoTrans, CblasConjTrans, alpha, param->do_a[i], param->do_a[i], beta, param->do_aad[i]);

			param->do_aad_eb[i] = gsl_matrix_complex_alloc(ns, ns);
		}
		if (param->do_g1[i] != 0.0)
		{
        		param->do_ada[i] = gsl_matrix_complex_alloc(ns, ns);
	        	gsl_blas_zgemm(CblasConjTrans, CblasNoTrans, alpha, param->do_a[i], param->do_a[i], beta, param->do_ada[i]);

			param->do_ada_eb[i] = gsl_matrix_complex_alloc(ns, ns);
		}
	}
	

	/* --- Evolve the system with an ODE solver --- */
	
	odeint_complex_rk_adaptive(qs, param, rho0, lme_ib_func_drho_dt, rho_cb_func);
	
	
	/* --- free matrices --- */
	for (i = 0; i < param->do_n; i++)
	{
		gsl_matrix_complex_free(param->do_ad_eb[i]);
		gsl_matrix_complex_free(param->do_a_eb[i]);

		if (param->do_aad[i] != NULL)
		{
			gsl_matrix_complex_free(param->do_aad[i]);
			gsl_matrix_complex_free(param->do_aad_eb[i]);
		}

		if (param->do_ada[i] != NULL)
		{
			gsl_matrix_complex_free(param->do_ada[i]);
			gsl_matrix_complex_free(param->do_ada_eb[i]);
		}
	}
	
	gsl_matrix_complex_free(param->ws1);
	gsl_matrix_complex_free(param->ws2);
	gsl_matrix_complex_free(param->op_tmp);

	return 0;
}


/* ---------------------------------------------------------------------------------------- 
 * Evolve the density matrix rho0 using non-dissipative dynamics and time-dependent
 * hamiltonian. 
 * param->Ti = start time
 * param->Tf = end time
 *
 *
 */
int
he_func_drho_dt(gsl_matrix_complex *rho_t, gsl_matrix_complex *drho_dt, double t,
                 quantum_system *qs, solver_param *p)
{
	gsl_complex z, alpha, beta;

	//hamiltonian_t(p->H_t, p->H0, p->H1, t, p);
	p->Ht_func(p->H_t, p->H0, p->H1, t, p);
	
	if (t==0.0)
    {
        printf("rho =\n");
    	misc_matrix_complex_print(rho_t, rho_t->size1);
	}
	
	// drho_dt = -i ( H * rho - rho * H )
    GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
    GSL_SET_COMPLEX(&beta,  0.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, rho_t, p->H_t, beta, drho_dt);
    GSL_SET_COMPLEX(&beta, -1.0, 0.0);
	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, p->H_t, rho_t, beta, drho_dt);
	GSL_SET_COMPLEX(&z, 0.0, -1.0);
	gsl_matrix_complex_scale(drho_dt, z);

	if (t==0.0)
    {
        printf("dH/dt =\n");
	    misc_matrix_complex_print(drho_dt, drho_dt->size1);
	}
	return 0;
}

int
dm_evolve_he_t(quantum_system *qs, gsl_matrix_complex *rho0,
		        solver_param *param, rho_cb_func_t rho_cb_func)
{
	int ns;

	//printf("DEBUG: starting dm_evolve_he_t (Lindblad-type Master Eqaution)\n");

        ns = quantum_system_nstates(qs);

	/* --- calculate hamiltonians --- */
	
	param->H0_func(param->H0, qs, param);
	param->H1_func(param->H1, qs, param);

	/* --- Evolve the system with an ODE solver --- */
	odeint_complex_rk(qs, param, rho0, he_func_drho_dt, rho_cb_func);
	
//	gsl_matrix_complex_free(param->H0);
//	gsl_matrix_complex_free(param->H1);
//	gsl_matrix_complex_free(param->H_t);

	return 0;
}



/* ---------------------------------------------------------------------------------------- 
 * Calculate the evolution operator:
 *
 *
 *
 */
int
dm_calculate_propagator(quantum_system *qs, solver_param *param, rho_cb_func_t rho_cb_func, gsl_matrix_complex *U)
{
	gsl_matrix_complex *rho0;
	
	int a, b, i, j, qsn = quantum_system_nstates(qs);

	rho0 = gsl_matrix_complex_alloc(qsn, qsn);
 	gsl_matrix_complex_set_zero(U);
	
//	printf("dm_calculate_propagator: debug: starting the evaluation of the propagator... \n");

	for (i = 0; i < qsn*qsn; i++)
	{
		//
		// setup (non-physical) initial density matrix
		//
	 	gsl_matrix_complex_set_zero(rho0);
		gsl_matrix_complex_set(rho0, ss_idx_minor_ms(qsn, i), ss_idx_minor_ls(qsn, i), GSL_COMPLEX_ONE);

//		printf("using the initial density matrix:\n");
//	        misc_matrix_complex_print_real(rho0, qsn);

		//
		// evolve
		//

		if (dm_evolve_lme_t(qs, rho0, param, rho_cb_func) != 0)
//		if (dm_evolve_lme_ib_t(qs, rho0, param, rho_cb_func) != 0) // for Fig 9
		{
			fprintf(stderr, "dm_calculate_propagator: Evolution of the quantum system failed.\n");
			return -1;
		}

//		printf("resulting density matrix:\n");
//	        misc_matrix_complex_print_real(param->rho_f, qsn);

		//
		// copy the resulting density matrix into U
		//		
		for (a = 0; a < qsn; a++)
		{
			for (b = 0; b < qsn; b++)
			{
				j = ss_idx_major(qsn, a, b);	
				gsl_matrix_complex_set(U, j, i, gsl_matrix_complex_get(param->rho_f, a, b));	
			}
		}

//		printf("current value of the propagator (real):\n");
//	        misc_matrix_complex_print_real(U, qsn*qsn);
//		printf("current value of the propagator (imag):\n");
//	        misc_matrix_complex_print_imag(U, qsn*qsn);
	}	
	
//	printf("dm_calculate_propagator: debug: done \n");

	gsl_matrix_complex_free(rho0);

	return 0;
}






































