/*
 * Functions for generating matrix representations of quantum operators.  
 * 
 * Robert Johansson <robert@riken.jp>
 *
 * $Id: operators.c,v 1.3 2008-09-11 10:00:57 rob Exp $
 */

#include <stdio.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

#include "quantum_system.h"
#include "density_matrix.h"
#include "operators.h"
#include "gsl_ext.h"

/* -----------------------------------------------------------------------------------
 * Spin 1/2 (qubit) operators:
 *
 */

/*
 * Generate the unit-matrix for the sub-system k.
 *
 */
int 
operator_unit(gsl_matrix_complex *u, quantum_system *qs, int k)
{
	int ns;
	
	if (k == -1)
		ns = quantum_system_nstates(qs);
	else
		ns = qs->nstates[k];

	gsl_matrix_complex_set_identity(u);
	
	return 0;
}

/*
 * Generate the sigma_z operator for the n:th subsystem.
 *
 */
int 
operator_sigma_z(gsl_matrix_complex *sz, quantum_system *qs, int n)
{
	quantum_system_state_vector_t qsv;
	int ns, i;

	if (qs == NULL)
	{
		fprintf(stderr, "operator_sigma_z called with NULL-valued qs argument\n");
		return -1;
	}

	if (qs->nstates[n] != 2)
	{
		fprintf(stderr, "The sigma_z operator cannot be generated for systems that are not TLSs.\n");
		return -1;
	}

	ns = quantum_system_nstates(qs);

	gsl_matrix_complex_set_zero(sz);

	for (i = 0; i < ns; i++) 
	{
		quantum_system_state_vector(qs, i, qsv);
		if (qsv[n] == 0)
			gsl_matrix_complex_set(sz, i, i, GSL_COMPLEX_ONE);
		else 
			gsl_matrix_complex_set(sz, i, i, GSL_COMPLEX_NEGONE);
	}

	return 0;
}


/*
 * Generate the sigma_x operator for the n:th subsystem.
 *
 */
int
operator_sigma_x(gsl_matrix_complex *sx, quantum_system *qs, int n)
{
	quantum_system_state_vector_t qsv;
	int ns, i, j;

	if (qs == NULL)
	{
		fprintf(stderr, "operator_sigma_x called with NULL-valued qs argument\n");
		return -1;
	}

	if (qs->nstates[n] != 2)
	{
		fprintf(stderr, "The sigma_x operator cannot be generated for systems that are not TLSs.\n");
		return -1;
	}

	ns = quantum_system_nstates(qs);

	gsl_matrix_complex_set_zero(sx);

	for (i = 0; i < ns; i++) 
	{
		quantum_system_state_vector(qs, i, qsv);

		if (qsv[n] == 1)
			qsv[n] = 0; 
		else
			qsv[n] = 1; 
		
		j = quantum_system_state_number(qs, qsv);
			
		gsl_matrix_complex_set(sx, i, j, GSL_COMPLEX_ONE);
	}

	return 0;
}

/*
 * Generate the sigma_y operator for the n:th subsystem.
 *
 */
int 
operator_sigma_y(gsl_matrix_complex *sy, quantum_system *qs, int n)
{
	quantum_system_state_vector_t qsv;
	int ns, i, j;
	gsl_complex z;

	if (qs == NULL)
	{
		fprintf(stderr, "operator_sigma_y called with NULL-valued qs argument\n");
		return -1;
	}

	if (qs->nstates[n] != 2)
	{
		fprintf(stderr, "The sigma_y operator cannot be generated for systems that are not TLSs.\n");
		return -1;
	}

	ns = quantum_system_nstates(qs);

	sy = gsl_matrix_complex_alloc(ns, ns);

	for (i = 0; i < ns; i++) 
	{
		quantum_system_state_vector(qs, i, qsv);

		if (qsv[n] == 1)
		{
			qsv[n] = 0; 
			GSL_SET_COMPLEX(&z, 0.0, 1.0);
		}
		else
		{
			qsv[n] = 1; 
			GSL_SET_COMPLEX(&z, 0.0, -1.0);
		}
		
		j = quantum_system_state_number(qs, qsv);
			
		gsl_matrix_complex_set(sy, i, j, z);
	}

	return 0;
}

/* -------------------------------------------------------------------------------
 * Harmonic oscillator operators:
 *
 */

/*
 * Generate the lowering operator for an harmonic oscillator that is
 * sub-system n: a
 *
 */
int
operator_ho_lowering(gsl_matrix_complex *a, quantum_system *qs, int n, int offset)
{
	density_matrix_list_t dm_list;
	int i, j;	

	density_matrix_list_init(&dm_list);
	
	for (i = 0; i < qs->nsubsys; i++)
	{
		gsl_matrix_complex *m;
		m = gsl_matrix_complex_alloc(qs->nstates[i], qs->nstates[i]);

		if (i != n)
		{
			operator_unit(m, qs, i);
		}
		else
		{
			gsl_matrix_complex_set_zero(m);
			for (j = 1; j < qs->nstates[i]; j++)
			{
				gsl_complex z;
				GSL_SET_COMPLEX(&z, sqrt(j+offset), 0);
				gsl_matrix_complex_set(m, j-1, j, z);
			}
		}

		density_matrix_list_append(&dm_list, m);
	}
	density_matrix_combine(a, qs, &dm_list);
	density_matrix_list_free_entries(&dm_list);
	
	return 0; 
}



/*
 * Generate the lowering operator for an harmonic oscillator that is
 * sub-system n: a dagger
 *
 */
int
operator_ho_raising(gsl_matrix_complex *a, quantum_system *qs, int n, int offset)
{
	density_matrix_list_t dm_list;
	int i, j;	

	density_matrix_list_init(&dm_list);
	
	for (i = 0; i < qs->nsubsys; i++)
	{
		gsl_matrix_complex *m;

		m = gsl_matrix_complex_alloc(qs->nstates[i], qs->nstates[i]);

		if (i != n)
		{
			operator_unit(m, qs, i);
		}
		else
		{
			gsl_matrix_complex_set_zero(m);
			for (j = 1; j < qs->nstates[i]; j++)
			{
				gsl_complex z;
                                GSL_SET_COMPLEX(&z, sqrt(j+offset), 0);
				gsl_matrix_complex_set(m, j, j-1, z);
			}
		}

		density_matrix_list_append(&dm_list, m);
	}
	density_matrix_combine(a, qs, &dm_list);
	density_matrix_list_free_entries(&dm_list);
	
	return 0; 
}

/*
 * Calculate the N = ad a operator for the nth sub-system.
 *
 */
int
operator_ho_N(gsl_matrix_complex *n_op, quantum_system *qs, int n, int offset)
{
        density_matrix_list_t dm_list;
        int i, j;

        density_matrix_list_init(&dm_list);

        for (i = 0; i < qs->nsubsys; i++)
        {
                gsl_matrix_complex *m = gsl_matrix_complex_alloc(qs->nstates[i], qs->nstates[i]);

                if (i != n)
                {
                        operator_unit(m, qs, i);
                }
                else
                {
                        gsl_matrix_complex_set_zero(m);
                        for (j = 0; j < qs->nstates[i]; j++)
                        {
                                gsl_complex z;
                                GSL_SET_COMPLEX(&z, j+offset, 0);
                                gsl_matrix_complex_set(m, j, j, z);
                        }
                }

                density_matrix_list_append(&dm_list, m);
        }
        density_matrix_combine(n_op, qs, &dm_list);
        density_matrix_list_free_entries(&dm_list);

        return 0;
//	
//	gsl_matrix_complex *ad, *a, *N_op;
//	gsl_complex alpha, beta;
//	int ns, i;
//
//	ns = quantum_system_nstates(qs);
//	
//	ad = operator_ho_raising(qs, n, offset);
//	a  = operator_ho_lowering(qs, n, offset);
//
//	N_op = gsl_matrix_complex_alloc(ns, ns);
//	
//	GSL_SET_COMPLEX(&alpha, 1.0, 0.0);
//	GSL_SET_COMPLEX(&beta, 0.0, 0.0);
//	gsl_blas_zgemm(CblasNoTrans, CblasNoTrans, alpha, ad, a, beta, N_op);
//
//	//for (i = 0; i < ns; i++)
//	//{
//	//	GSL_SET_COMPLEX(&alpha, i+offset, 0.0);
//	//	gsl_matrix_complex_set(N_op, i, i, alpha);
//	//}
//	
//	gsl_matrix_complex_free(ad);
//	gsl_matrix_complex_free(a);
//
//	return N_op;
}

/* -------------------------------------------------------------------------------
 * Generic creation/annihilation operators:
 *
 */
int
operator_project(gsl_matrix_complex *op, quantum_system *qs, int n, int si, int sj)
{
        density_matrix_list_t dm_list;
        int i, j;

        density_matrix_list_init(&dm_list);

        for (i = 0; i < qs->nsubsys; i++)
        {
                gsl_matrix_complex *m = gsl_matrix_complex_alloc(qs->nstates[n], qs->nstates[n]);

                if (i != n)
                {
                        operator_unit(m, qs, i);
                }
                else
                {
                        gsl_matrix_complex_set_zero(m);
			gsl_matrix_complex_set(m, sj, si, GSL_COMPLEX_ONE);
                }

                density_matrix_list_append(&dm_list, m);
        }
        density_matrix_combine(op, qs, &dm_list);
        density_matrix_list_free_entries(&dm_list);

        return 0;
}

/* ------------------------------------------------------------------------------------------
 * 3-level system operators:
 *
 *
 */

/*
 * Generate the 3LS sigma_z operator for the n:th subsystem.
 *
 */
int
operator_3ls_sigma_z(gsl_matrix_complex *sz, quantum_system *qs, int n)
{
	quantum_system_state_vector_t qsv;
	int ns, i;
	
	if (qs == NULL)
	{
		fprintf(stderr, "operator_sigma_z called with NULL-valued qs argument\n");
		return -1;
	}

	if (qs->nstates[n] != 3)
	{
		fprintf(stderr, "This sigma_z operator cannot be generated for systems that are not 3LS.\n");
		return -1;
	}

	ns = quantum_system_nstates(qs);

	sz = gsl_matrix_complex_alloc(ns, ns);

	for (i = 0; i < ns; i++) 
	{
		quantum_system_state_vector(qs, i, qsv);
		if (qsv[n] == 0)
			gsl_matrix_complex_set(sz, i, i, GSL_COMPLEX_ONE);
		else 
			gsl_matrix_complex_set(sz, i, i, GSL_COMPLEX_NEGONE);
	}

	return 0;
}


/*
 * Generate the 3LS sigma_x operator for the n:th subsystem.
 *
 */
int
operator_3ls_sigma_x(gsl_matrix_complex *sx, quantum_system *qs, int n)
{
	quantum_system_state_vector_t qsv;
	int ns, i, j;

	if (qs == NULL)
	{
		fprintf(stderr, "operator_sigma_x called with NULL-valued qs argument\n");
		return -1;
	}

	if (qs->nstates[n] != 3)
	{
		fprintf(stderr, "This sigma_x operator cannot be generated for systems that are a 3LS.\n");
		return -1;
	}

	ns = quantum_system_nstates(qs);

	gsl_matrix_complex_set_zero(sx);

	for (i = 0; i < ns; i++) 
	{
		quantum_system_state_vector(qs, i, qsv);

		if (qsv[n] != 2)
		{
			if (qsv[n] == 1)
				qsv[n] = 0; 
			else 
				qsv[n] = 1; 
		
			j = quantum_system_state_number(qs, qsv);
			
			gsl_matrix_complex_set(sx, i, j, GSL_COMPLEX_ONE);
		}
	}

	return 0;
}



/* --------------------------------------------------------------------------------
 * Angular momentum operators:
 *
 */

/*
 * Generate the sigma_z operator for the n:th subsystem.
 *
 */
int
operator_Jz(gsl_matrix_complex *Jz, quantum_system *qs, int n, double j)
{
        quantum_system_state_vector_t qsv;
        int ns, i;
	double m;
	int N = (int)(2 * j + 1);

        if (qs == NULL)
        {
                fprintf(stderr, "operator_Jz called with NULL-valued qs argument\n");
                return -1;
        }

        if (qs->nstates[n] != N)
        {
                fprintf(stderr, "The sigma_Jz operator needs %d quantum states (given %d states).\n", N, qs->nstates[n]);
                return -1;
        }

        ns = quantum_system_nstates(qs);
        gsl_matrix_complex_set_zero(Jz);

        for (i = 0; i < ns; i++)
        {
		gsl_complex z;

                quantum_system_state_vector(qs, i, qsv);

		m = j - qsv[n];		

		GSL_SET_COMPLEX(&z, m, 0.0);
		gsl_matrix_complex_set(Jz, i, i, z);
        }

        return 0;
}

/*
 * Generate matrix representation of the J+ operator
 * 
 */
int
operator_J_plus(gsl_matrix_complex *Jp, quantum_system *qs, int n, double jj)
{
        density_matrix_list_t dm_list;
        int i, j, N = (int)(2 * jj + 1);
	double m;

        if (qs->nstates[n] != N)
        {
                fprintf(stderr, "The sigma_Jp operator needs %d quantum states (given %d states).\n", N, qs->nstates[n]);
                return -1;
        }

        density_matrix_list_init(&dm_list);

        for (i = 0; i < qs->nsubsys; i++)
        {
		gsl_matrix_complex *M = gsl_matrix_complex_alloc(qs->nstates[n], qs->nstates[n]);

                if (i != n)
                {
                        operator_unit(M, qs, i);
                }
                else
                {
                        gsl_matrix_complex_set_zero(M);

                        for (j = 1; j < qs->nstates[n]; j++)
                        {
                                gsl_complex z;

				m = jj - j;

				//printf("operator_J_plus: DEBUG: idx = %d, j = %f, m = %f, sqrt(..) = %f\n",
				//	j, jj, m, sqrt(jj*(jj+1) - m*(m+1)));

                                GSL_SET_COMPLEX(&z, sqrt(jj*(jj+1) - m*(m+1)), 0);

                                gsl_matrix_complex_set(M, j-1, j, z);
                        }
                }

                density_matrix_list_append(&dm_list, M);
        }
        density_matrix_combine(Jp, qs, &dm_list);
        density_matrix_list_free_entries(&dm_list);

        return 0;
}

/*
 * Generate matrix representation of the J- operator
 *
 */
int
operator_J_minus(gsl_matrix_complex *Jm, quantum_system *qs, int n, double jj)
{
        density_matrix_list_t dm_list;
        int i, j, N = (int)(2 * jj + 1);
        double m;

        if (qs->nstates[n] != N)
        {
                fprintf(stderr, "The sigma_Jp operator needs %d quantum states (given %d states).\n", N, qs->nstates[n]);
                return -1;
        }

        density_matrix_list_init(&dm_list);

        for (i = 0; i < qs->nsubsys; i++)
        {
	        gsl_matrix_complex *M = gsl_matrix_complex_alloc(qs->nstates[n], qs->nstates[n]);

                if (i != n)
                {
                        operator_unit(M, qs, i);
                }
                else
                {
                        gsl_matrix_complex_set_zero(M);

                        for (j = 0; j < qs->nstates[n]-1; j++)
                        {
                                gsl_complex z;

                                m = jj - j;

                                GSL_SET_COMPLEX(&z, sqrt(jj*(jj+1) - m*(m-1)), 0);
                                //GSL_SET_COMPLEX(&z, m, 0);

                                gsl_matrix_complex_set(M, j+1, j, z);
                        }
                }

                density_matrix_list_append(&dm_list, M);
        }
        density_matrix_combine(Jm, qs, &dm_list);
        density_matrix_list_free_entries(&dm_list);

        return 0;
}



/* ------------------------------------------------------------------------------------
 * Computational routines:
 *
 */

/*
 * Calculate the trace of a matrix (operator). Used for calculating expectation
 * values and two-time correlation functions.
 *
 * XXX: SHOULD BE OPTIMIZED
 */
gsl_complex
operator_trace(gsl_matrix_complex *op)
{
        gsl_complex tr;
        int i;

        GSL_SET_COMPLEX(&tr, 0.0, 0.0);
        for (i = 0; i < op->size1; i++)
                tr = gsl_complex_add(tr, gsl_matrix_complex_get(op, i, i));

        return tr;
}



